# DRI configuration GUI: simplified UI components

# Copyright (C) 2003-2006  Felix Kuehling

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Contact: http://fxk.de.vu/

import os
import dri
import pygtk
pygtk.require ("2.0")
import gtk
import gobject

import driconf_commonui
import driconf_complexui
commonui = driconf_commonui    # short cut
complexui = driconf_complexui

from driconf_commonui import _, lang

def isUserConfig(config):
    return config.fileName.startswith(os.environ["HOME"])

def getUserConfig(configList):
    userConfigs = [config for config in configList if isUserConfig(config)]
    if not userConfigs:
        return None
    else:
        return userConfigs[0]

def genNormalDeviceConfigs (configList, dpy):
    """ Generate a list of normalized device configurations.

    One device configuration for each installed device. Each contains
    a default application configuration that explicitly sets all
    options to their default values in order to override values in
    previous less specific device sections. Then it appends
    application configurations for all application configurations in
    all configuration files that affect the respective device. The
    result is one device configuration per existing device that
    summarizes the entire configuration of existing devices and
    overrides any previous settings.

    If there is no user configuration file, an empty list is returned. """
    userConfig = getUserConfig(configList)
    if not userConfig:
        return []
    screens = [screen for screen in dpy.screens if screen]
    deviceConfigs = []
    # Create one device configuration for each installed device on this display
    for screen in screens:
        if screen == None:
            continue
        driver = screen.driver
        deviceConfig = dri.DeviceConfig(userConfig, str(screen.num), driver.name)
        defaultApp = dri.AppConfig(deviceConfig, "Default")
        deviceConfig.apps.append(defaultApp)
        for sect in driver.optSections:
            for opt in sect.options.values():
                defaultApp.options[opt.name] = dri.ValueToStr(opt.default, opt.type)
        deviceConfig.isNormalized = True
        deviceConfigs.append(deviceConfig)
    for config in configList:
        configIsUser = isUserConfig(config)
        for device in config.devices:
            # Determine all installed devices affected by this device-section
            # in the original configuration file
            curDeviceConfigs = [deviceConfig for deviceConfig in deviceConfigs
                                if (device.screen == None or
                                    device.screen == deviceConfig.screen)
                                and (device.driver == None or
                                     device.driver == deviceConfig.driver)]
            for curDevice in curDeviceConfigs:
                driver = dpy.getScreen(int(curDevice.screen)).driver
                for app in device.apps:
                    # Determine all applications on this device affected by
                    # this application section in the original config file.
                    # It should be one at most. Create a new application
                    # configuration if needed.
                    curApps = [curApp for curApp in curDevice.apps
                               if app.executable == curApp.executable]
                    assert len(curApps) <= 1
                    if curApps:
                        curApp = curApps[0]
                    else:
                        curApp = dri.AppConfig(curDevice, app.name, app.executable)
                        curDevice.apps.append(curApp)
                    # Update all option settings. Non-existing options
                    # or invalid values are only considered in
                    # redundant device sections.
                    for opt,value in app.options.items():
                        isValid = False
                        if configIsUser and \
                               device.screen != None and device.driver != None:
                            isValid = True
                        else:
                            optInfo = driver.getOptInfo(opt)
                            if optInfo and optInfo.validate(value):
                                isValid = True
                        if isValid:
                            curApp.options[opt] = value
    return deviceConfigs

def removeRedundantDevices (config, normalDeviceConfigs, onlyTest = False):
    """ Remove device configurations that are redundant ...

    ... after appending normalized device configurations. If onlyTest
    is True, the configuration file is not modified and this function
    returns True iff there are redundant device sections. Otherwise
    False is returned. """
    screens = [device.screen for device in normalDeviceConfigs]
    # Iterate over a copy of the device list, so that devices can be
    # removed safely.
    for device in config.devices[:]:
        if not (hasattr(device, "isNormalized") and device.isNormalized) and \
               device.screen != None and device.driver != None:
            # See if there is a normalized device configuration for
            # this device. In that case this section is redundant.
            try:
                i = screens.index(device.screen)
            except ValueError:
                pass
            else:
                if normalDeviceConfigs[i].driver == device.driver:
                    if onlyTest:
                        return True
                    # Remove redundant device section
                    config.devices.remove(device)
                    config.isModified = True
    return False

def isRedundant (configList, dpy, normalDeviceConfigs = None):
    """ Check if the user configuration is redundant.

    Returns True iff there is a user configuration file that would
    contain redundant device configurations after appending
    normalDeviceConfigs. """
    userConfig = getUserConfig(configList)
    if not userConfig:
        return False
    if normalDeviceConfigs == None:
        normalDeviceConfigs = genNormalDeviceConfigs (configList, dpy)
    return removeRedundantDevices (userConfigs, normalDeviceConfigs,
                                   onlyTest=True)

def isNormalized(configList, dpy, normalDeviceConfigs = None):
    """ Check if the user configuration file is normalized ...

    ... in a set of configuration files, that is if the user
    configuration file would be the same (except for names) after
    normalization. If the user configuration file is normalized, a
    list of existing normalized device configurations is returned. If
    there is no user configuration file, an empty list is
    returned. Otherwise, if there is a user configuration file that is
    not normalized, this function returns None."""
    userConfig = getUserConfig(configList)
    if not userConfig:
        return []
    userDevs = userConfig.devices
    # Find a consistent list of specific device configurations at the
    # end of the user configuration file.
    i = len(userDevs) - 1
    while i >= 0 and userDevs[i].screen != None and userDevs[i].driver != None:
        i = i - 1
    i = i + 1
    specificDevs = userDevs[i:]
    # Make sure there is at least one for each configurable screen.
    # If there are several the last one counts.
    screens = [screen for screen in dpy.screens if screen]
    screenDevs = [None for i in range(len(screens))]
    for device in specificDevs:
        screenNum = int(device.screen)
        if screenNum >= len(screenDevs):
            continue
        if screens[screenNum].driver.name == device.driver:
            screenDevs[screenNum] = device
    if [None for device in screenDevs if device == None]:
        return None  # There are unconfigured screens
    if normalDeviceConfigs == None:
        normalDeviceConfigs = genNormalDeviceConfigs (configList, dpy)
    # Compare existing normalized device configs with generated
    # ones. If they are equivalent, the configuration file is
    # normalized.
    for device,normalDev in zip (screenDevs,normalDeviceConfigs):
        # Check that the first executable is None and that each
        # executable is configured exactly once.
        executables = [app.executable for app in device.apps]
        if not executables or executables[0] != None:
            return None
        del executables[0]
        if [None for exe in executables if exe == None]:
            return None
        executables.sort()
        normalExes = [app.executable for app in normalDev.apps]
        del normalExes[0]
        normalExes.sort()
        if normalExes != executables:
            return None
        # Now check that each application contains the same option settings
        # as the generated normalized configuration
        for normalApp in normalDev.apps:
            app = [dApp for dApp in device.apps
                   if dApp.executable == normalApp.executable][0]
            if app.options != normalApp.options:
                return None
    # The configuration is normalized. Return the list of normalized device
    # configurations from the user configuration files.
    return screenDevs

def normalizeConfig(configList, dpy):
    """ Normalize the user configuration file (if it exists) ...

    ... by appending normalized device configurations for each
    installed device and removing redundant device configurations. If
    the user configuration file is already normalized, only existing
    normalized device configurations are marked as such and redundant
    device configurations are removed. """
    newDeviceConfigs = genNormalDeviceConfigs (configList, dpy)
    existingDeviceConfigs = isNormalized (configList, dpy, newDeviceConfigs)
    if not existingDeviceConfigs and not newDeviceConfigs:
        return []
    userConfig = getUserConfig(configList)
    if existingDeviceConfigs:
        # is already normalized, mark existing normalized device
        # configurations as such.
        for deviceConfig in existingDeviceConfigs:
            deviceConfig.isNormalized = True
        deviceConfigs = existingDeviceConfigs
    elif newDeviceConfigs:
        userConfig.devices.extend(newDeviceConfigs)
        userConfig.isModified = True
        deviceConfigs = newDeviceConfigs
    # Remove redundant device configurations from the user
    # configuration file
    removeRedundantDevices (userConfig, deviceConfigs)
    return deviceConfigs

def lineWrap (string, chars=30):
    head = ""
    tail = string
    while len(tail):
        if len(tail) <= chars:
            return head + tail
        else:
            i = chars
            while i >= chars/3:
                if tail[i] == ' ':
                    j = i + 1
                    break
                elif tail[i] == '-':
                    i = i + 1
                    j = i
                    break
                i = i - 1
            if i < chars/3:
                i = chars
            head, tail = head + tail[:i] + '\n', tail[j:]
    return head

class AppDialog (gtk.Dialog):
    def __init__ (self, title, parent, app=None):
        gtk.Dialog.__init__(self, title, parent,
                            gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
                            ("gtk-ok", gtk.RESPONSE_OK,
                             "gtk-cancel", gtk.RESPONSE_CANCEL))
        self.set_resizable(False)
        table = gtk.Table(3, 2)
        nameLabel = gtk.Label(_("Application Name"))
        nameLabel.show()
        table.attach(nameLabel, 0, 1, 0, 1, 0, gtk.EXPAND, 10, 5)
        self.nameEntry = gtk.Entry()
        if app:
            self.nameEntry.set_text(app.name)
        self.nameEntry.connect("activate",
                               lambda widget: self.response(gtk.RESPONSE_OK))
        self.nameEntry.show()
        table.attach(self.nameEntry, 1, 2, 0, 1,
                     gtk.EXPAND|gtk.FILL, gtk.EXPAND, 10, 5)

        execLabel = gtk.Label(_("Executable Name"))
        execLabel.show()
        table.attach(execLabel, 0, 1, 1, 2, 0, gtk.EXPAND, 10, 5)
        self.execEntry = gtk.Entry()
        if app:
            self.execEntry.set_text(app.executable)
        self.execEntry.connect("activate",
                               lambda widget: self.response(gtk.RESPONSE_OK))
        self.execEntry.show()
        table.attach(self.execEntry, 1, 2, 1, 2,
                     gtk.EXPAND|gtk.FILL, gtk.EXPAND, 10, 5)
        hBox = gtk.HBox(spacing=10)
        infoImageVBox = gtk.VBox()
        infoImage = commonui.StockImage(gtk.STOCK_DIALOG_INFO,
                                        gtk.ICON_SIZE_DIALOG)
        infoImage.show()
        infoImageVBox.pack_start(infoImage, False, False, 0)
        infoImageVBox.show()
        hBox.pack_start(infoImageVBox, False, False, 0)
        infoLabel = gtk.Label(_(
            "The executable name is important for identifying the "
            "application. If you get it wrong, your settings will not "
            "apply. Beware that some applications are started by a "
            "shell script, that has a different name than the real "
            "executable."))
        # TODO: Add a small database of known applications with their
        # executable names that can be picked from a menu.
        infoLabel.set_line_wrap (True)
        infoLabel.show()
        hBox.pack_start(infoLabel, False, False, 0)
        hBox.show()
        table.attach (hBox, 0, 2, 2, 3,
                      gtk.EXPAND|gtk.FILL, gtk.EXPAND, 10, 5)
        table.show()
        self.vbox.pack_start(table, True, True, 5)
        self.show()
        self.nameEntry.grab_focus()

    def getName (self):
        return self.nameEntry.get_text()

    def getExecutable (self):
        return self.execEntry.get_text()

class AppPage (gtk.ScrolledWindow):
    def __init__ (self, driver, app):
        """ Constructor. """
        gtk.ScrolledWindow.__init__ (self)
        self.set_policy (gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        self.driver = driver
        self.app = app
        self.tooltips = gtk.Tooltips()
        self.table = None
        self.refreshOptions()

    def refreshOptions (self):
        if self.table:
            self.remove(self.get_child())
        self.optLines = []
        self.table = gtk.Table(len(self.app.options)+1, 3)
        self.add_with_viewport(self.table)
        self.optionTree = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_INT)
        i = 0
        sectI = 0
        for sect in self.driver.optSections:
            sectIter = self.optionTree.append(None, [
                lineWrap(sect.getDesc([lang])), sectI, -1])
            sectHasOpts = False
            optI = 0
            for opt in sect.optList:
                if self.app.options.has_key(opt.name):
                    self.optLines.append(
                        commonui.OptionLine(self, i, opt, True, True))
                    i = i + 1
                else:
                    self.optionTree.append(sectIter, [
                        lineWrap(opt.getDesc([lang]).text), sectI, optI])
                    sectHasOpts = True
                optI = optI + 1
            if not sectHasOpts:
                self.optionTree.remove(sectIter)
            sectI = sectI + 1
        if len(self.optionTree) > 0:
            addLabel = commonui.WrappingDummyCheckButton(_("Add setting"),
                                                         width=200)
            addLabel.show()
            self.table.attach(addLabel, 0, 1, i, i+1, gtk.EXPAND|gtk.FILL, 0, 5, 5)
            addCombo = gtk.ComboBox(self.optionTree)
            addCombo.connect("changed", self.addOption)
            cell = gtk.CellRendererText()
            addCombo.pack_start(cell, True)
            addCombo.add_attribute(cell, 'text', 0)
            addCombo.show()
            self.table.attach(addCombo, 1, 2, i, i+1, gtk.FILL, 0, 5, 5)
        self.table.show()

    def optionModified (self, optLine):
        self.app.modified(self.app)

    def removeOption (self, optLine, opt):
        del self.app.options[opt.name]
        self.refreshOptions()
        self.app.modified(self.app)

    def addOption (self, combo):
        activeIter = combo.get_active_iter()
        if not activeIter:
            # Got triggered by the set_active(-1) below.
            return
        sectI = self.optionTree.get_value(activeIter, 1)
        optI  = self.optionTree.get_value(activeIter, 2)
        if optI < 0:
            combo.set_active(-1)
            return
        opt = self.driver.optSections[sectI].optList[optI]
        self.app.options[opt.name] = dri.ValueToStr(opt.default, opt.type)
        self.refreshOptions()
        self.app.modified(self.app)
    
    def doValidate (self):
        pass

    def commit (self):
        for optLine in self.optLines:
            name = optLine.opt.name
            value = optLine.getValue()
            if value == None and self.app.options.has_key(name):
                del self.app.options[name]
            elif value != None:
                self.app.options[name] = value

class MainWindow (gtk.Window):
    def __init__ (self, configList):
        gtk.Window.__init__(self)
        self.set_title(_("Direct Rendering Preferences"))
        self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_DIALOG)
        self.set_border_width(10)
        self.connect("destroy", lambda dummy: gtk.main_quit())
        self.connect("delete_event", self.exitHandler)
        self.configList = configList # Remember for switching to expert mode
        self.userConfig = getUserConfig(configList)
        self.screens = [screen for screen in commonui.dpy.screens if screen]
        self.vbox = gtk.VBox(spacing=10)
        if len(self.screens) > 1:
            self.deviceCombo = gtk.combo_box_new_text()
            for screen in self.screens:
                if screen.glxInfo:
                    self.deviceCombo.append_text(_("Screen") + " %d: %s (%s)" % (
                        screen.num, screen.glxInfo.renderer, screen.glxInfo.vendor))
                else:
                    self.deviceCombo.append_text(_("Screen") + " %d: %s" % (
                        screen.num, screen.driver.name.capitalize()))
            self.deviceCombo.set_active(0)
            self.deviceCombo.connect("changed", self.changeDevice)
            self.deviceCombo.show()
            self.vbox.pack_start(self.deviceCombo, False, False, 0)
        else:
            screen = self.screens[0]
            if screen.glxInfo:
                text = "%s (%s)" % (
                    screen.glxInfo.renderer, screen.glxInfo.vendor)
            else:
                text = _("Screen") + " %d: %s" % (
                    screen.num, screen.driver.name.capitalize())
            deviceHBox = gtk.HBox()
            deviceLabel = gtk.Label()
            deviceLabel.set_justify(gtk.JUSTIFY_LEFT)
            deviceLabel.set_markup("<b>" + commonui.escapeMarkup(text) + "</b>")
            deviceLabel.show()
            deviceHBox.pack_start(deviceLabel, False, False, 0)
            deviceHBox.show()
            self.vbox.pack_start(deviceHBox, False, False, 0)
        buttonBox = gtk.HButtonBox()
        buttonBox.set_layout(gtk.BUTTONBOX_END)
        expertButton = gtk.Button()
        expertHBox = gtk.HBox()
        expertImage = commonui.StockImage("gtk-jump-to", gtk.ICON_SIZE_BUTTON)
        expertImage.show()
        expertHBox.pack_start(expertImage)
        expertLabel = gtk.Label(_("Expert Mode"))
        expertLabel.show()
        expertHBox.pack_start(expertLabel)
        expertHBox.show()
        expertButton.add(expertHBox)
        expertButton.connect("clicked", self.expertHandler)
        expertButton.show()
        buttonBox.add(expertButton)
        closeButton = gtk.Button(stock="gtk-close")
        closeButton.connect("clicked", lambda dummy: gtk.main_quit())
        closeButton.show()
        buttonBox.add(closeButton)
        aboutButton = gtk.Button(stock="gtk-about")
        aboutButton.connect("clicked", self.aboutHandler)
        aboutButton.show()
        buttonBox.add(aboutButton)
        buttonBox.set_child_secondary(aboutButton, True)
        buttonBox.show()
        self.vbox.pack_end(buttonBox, False, False, 0)
        self.expander = gtk.Expander(
            "<b>" + commonui.escapeMarkup(_("Application settings")) + "</b>")
        self.expander.set_use_markup(True)
        self.expander.connect("activate", self.expanderChanged)
        self.expander.show()
        self.vbox.pack_end(self.expander, False, True, 0)
        self.expanderVBox = gtk.VBox(spacing=10)
        self.appButtonBox = gtk.HBox()
        self.appRemoveButton = gtk.Button(stock="gtk-remove")
        self.appRemoveButton.connect("clicked", self.removeApp)
        self.appRemoveButton.show()
        self.appButtonBox.pack_end(self.appRemoveButton, False, False, 0)
        addButton = gtk.Button(stock="gtk-add")
        addButton.connect("clicked", self.addApp)
        addButton.show()
        self.appButtonBox.pack_end(addButton, False, False, 0)
        self.appPropButton = gtk.Button(stock="gtk-properties")
        self.appPropButton.connect("clicked", self.appProperties)
        self.appPropButton.show()
        self.appButtonBox.pack_end(self.appPropButton, False, False, 0)
        self.appButtonBox.show()
        self.expanderVBox.pack_start(self.appButtonBox, False, False, 0)
        self.expanderVBox.show()
        self.expander.add(self.expanderVBox)
        self.notebook = None
        self.appCombo = None
        self.appNotebook = None
        self.appPage = None
        self.selectScreen(0)
        self.vbox.show()
        self.add(self.vbox)

    def selectScreen (self, n):
        self.curScreen = self.screens[n]
        # Find that device's configuration in the user config. Search
        # from the end, because that's where the normalized configs are.
        self.deviceConfig = None
        i = len(self.userConfig.devices)-1
        while i >= 0 and self.userConfig.devices[i].screen != None and \
              self.userConfig.devices[i].driver != None:
            if self.curScreen.num == int(self.userConfig.devices[i].screen) and \
               self.curScreen.driver.name == self.userConfig.devices[i].driver:
                self.deviceConfig = self.userConfig.devices[i]
                self.driver = self.curScreen.driver
                break
            i = i - 1
        # Sanity checks. These should all be true after normalization.
        assert(self.deviceConfig)
        assert(self.deviceConfig.isNormalized)
        assert(len(self.deviceConfig.apps) > 0 and
               self.deviceConfig.apps[0].executable == None)
        # Register modified callback
        self.deviceConfig.apps[0].modified = self.configModified
        # Build UI for the screen configuration
        if self.notebook:
            self.vbox.remove(self.notebook)
        self.notebook = gtk.Notebook()
        self.notebook.popup_enable()
        self.sectPages = []
        self.sectLabels = []
        unknownPage = commonui.UnknownSectionPage (self.driver,
                                                   self.deviceConfig.apps[0])
        if len(unknownPage.opts) > 0:
            unknownPage.show()
            unknownLabel = gtk.Label (_("Unknown options"))
            unknownLabel.show()
            self.notebook.append_page (unknownPage, unknownLabel)
            self.sectPages.append (unknownPage)
            self.sectLabels.append (unknownLabel)
        for sect in self.driver.optSections:
            sectPage = commonui.SectionPage (sect, self.deviceConfig.apps[0],
                                             True)
            sectPage.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
            sectPage.show()
            desc = sect.getDesc([lang])
            if desc:
                sectLabel = gtk.Label (desc)
                sectLabel.set_line_wrap (True)
            else:
                sectLabel = gtk.Label (_("(no description)"))
            sectLabel.show()
            self.notebook.append_page (sectPage, sectLabel)
            self.sectPages.append (sectPage)
            self.sectLabels.append (sectLabel)
        if len(self.sectLabels) > 0:
            style = self.sectLabels[0].get_style()
            self.default_normal_fg = style.fg[gtk.STATE_NORMAL].copy()
            self.default_active_fg = style.fg[gtk.STATE_ACTIVE].copy()
        self.validate()
        self.notebook.show()
        self.vbox.pack_start(self.notebook, True, True, 0)
        if self.appCombo:
            self.appButtonBox.remove(self.appCombo)
        self.appCombo = gtk.combo_box_new_text()
        for i in range(1,len(self.deviceConfig.apps)):
            self.appCombo.append_text(self.deviceConfig.apps[i].name)
        if len(self.deviceConfig.apps) > 1:
            self.appCombo.set_active(0)
            self.expander.set_expanded(True)
            self.vbox.set_child_packing(self.expander, True, True,
                                        0, gtk.PACK_END)
            self.appPropButton.set_sensitive(True)
            self.appRemoveButton.set_sensitive(True)
        else:
            self.expander.set_expanded(False)
            self.vbox.set_child_packing(self.expander, False, True,
                                        0, gtk.PACK_END)
            self.appPropButton.set_sensitive(False)
            self.appRemoveButton.set_sensitive(False)
        self.appCombo.connect("changed", self.changeApp)
        self.appCombo.show()
        self.appButtonBox.pack_start(self.appCombo, True, True, 0)
        if len(self.deviceConfig.apps) > 1:
            self.selectApp(self.deviceConfig.apps[1])
        else:
            self.selectApp(None)

    def selectApp (self, app):
        if self.appNotebook:
            self.expanderVBox.remove(self.appNotebook)
            self.appNotebook = None
            self.appPage = None
        elif self.appPage:
            self.expanderVBox.remove(self.appPage)
            self.appPage = None
        if not app:
            self.appPropButton.set_sensitive(False)
            self.appRemoveButton.set_sensitive(False)
            return
        self.appPropButton.set_sensitive(True)
        self.appRemoveButton.set_sensitive(True)
        app.modified = self.configModified
        unknownPage = commonui.UnknownSectionPage (self.driver,
                                                   app)
        if len(unknownPage.opts) > 0:
            self.appNotebook = gtk.Notebook()
            self.appNotebook.popup_enable()
            unknownPage.show()
            unknownLabel = gtk.Label (_("Unknown options"))
            unknownLabel.show()
            self.appNotebook.append_page (unknownPage, unknownLabel)
            self.appPage = AppPage (self.driver, app)
            self.appPage.show()
            appPageLabel = gtk.Label (_("Known options"))
            appPageLabel.show()
            self.appNotebook.append_page (self.appPage, appPageLabel)
            self.appNotebook.show()
            appWidget = self.appNotebook
        else:
            self.appPage = AppPage (self.driver, app)
            self.appPage.show()
            appWidget = self.appPage
        self.expanderVBox.pack_start (appWidget, True, True, 0)

    def validate (self):
        """ Validate the configuration.

        Labels of invalid section pages are highlighted. Returns whether
        there were invalid option values. """
        index = 0
        allValid = True
        for sectPage in self.sectPages:
            valid = sectPage.validate()
            if not valid:
                # strange, active and normal appear to be swapped :-/
                self.sectLabels[index].modify_fg (
                    gtk.STATE_NORMAL, gtk.gdk.Color (65535, 0, 0))
                self.sectLabels[index].modify_fg (
                    gtk.STATE_ACTIVE, gtk.gdk.Color (65535, 0, 0))
            else:
                self.sectLabels[index].modify_fg (
                    gtk.STATE_NORMAL, self.default_normal_fg)
                self.sectLabels[index].modify_fg (
                    gtk.STATE_ACTIVE, self.default_active_fg)
            allValid = allValid and valid
            index = index+1
        return allValid        

    def commit (self):
        for sectPage in self.sectPages:
            sectPage.commit()
        if self.appPage:
            self.appPage.commit()

    def changeDevice (self, combo):
        self.selectScreen(combo.get_active())

    def changeApp (self, combo):
        active = combo.get_active()
        if active >= 0:
            app = self.deviceConfig.apps[active+1]
        else:
            app = None
        if self.appPage:
            lastApp = self.appPage.app
        else:
            lastApp = None
        if app != lastApp:
            if lastApp:
                self.appPage.commit()
            self.selectApp(app)

    def expanderChanged (self, expander):
        # This signal handler seems to get called before the expander
        # state is changed. So the logic is reversed.
        if not self.expander.get_expanded():
            self.vbox.set_child_packing(self.expander, True, True,
                                        0, gtk.PACK_END)
        else:
            self.vbox.set_child_packing(self.expander, False, True,
                                        0, gtk.PACK_END)

    def checkAppProperties (self, dialog, name, executable, sameApp=None):
        errorStr = None
        if name == "" or executable == "":
            # Error message
            errorStr = _("You must enter both an application name and "
                         "an executable name.")
        else:
            for app in self.deviceConfig.apps:
                if app != sameApp and name == app.name:
                    errorStr = _("There exists an application "
                                 "configuration with the same name. "
                                 "Please enter a different name.")
                    break
                elif app != sameApp and executable == app.executable:
                    errorStr = _("There exists an application "
                                 "configuration for the same "
                                 "executable. You can't create multiple "
                                 "application configurations for the "
                                 "same executable.")
                    break
        if errorStr:
            dialog = gtk.MessageDialog(
                dialog, gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, errorStr)
            dialog.run()
            dialog.destroy()
            return False
        return True

    def addApp (self, button):
        dialog = AppDialog(_("Add Application"), self)
        done = False
        while not done:
            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                name = dialog.getName().strip()
                executable = dialog.getExecutable().strip()
                if self.checkAppProperties (dialog, name, executable):
                    app = dri.AppConfig(self.deviceConfig, name, executable)
                    self.deviceConfig.apps.append(app)
                    self.appCombo.append_text(name)
                    self.appCombo.set_active(len(self.deviceConfig.apps)-2)
                    self.configModified (self.deviceConfig)
                    done = True
            else:
                done = True
        dialog.destroy()

    def appProperties (self, button):
        if not self.appPage:
            return
        dialog = AppDialog(_("Application Properties"), self, self.appPage.app)
        done = False
        while not done:
            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                name = dialog.getName().strip()
                executable = dialog.getExecutable().strip()
                if self.checkAppProperties (dialog, name, executable,
                                            self.appPage.app):
                    i = self.deviceConfig.apps.index(self.appPage.app)
                    self.appCombo.remove_text(i-1)
                    self.appCombo.insert_text(i-1, name)
                    self.appCombo.set_active(i-1)
                    self.appPage.app.name = name
                    self.appPage.app.executable = executable
                    self.configModified (self.appPage.app)
                    done = True
            else:
                done = True
        dialog.destroy()

    def removeApp (self, button):
        if not self.appPage:
            return
        i = self.deviceConfig.apps.index(self.appPage.app)
        if i+1 < len(self.deviceConfig.apps):
            newI = i
            newApp = self.deviceConfig.apps[i+1]
        elif i > 1:
            newI = i-1
            newApp = self.deviceConfig.apps[i-1]
        else:
            newI = 0
            newApp = None
        del self.deviceConfig.apps[i]
        self.selectApp(newApp)
        self.appCombo.remove_text(i-1)
        self.appCombo.set_active(newI-1)
        self.configModified(self.deviceConfig)

    def exitHandler (self, widget, event=None):
        # Always ok to destroy the window
        return False

    def aboutHandler (self, widget):
        dialog = commonui.AboutDialog()
        dialog.show()
        dialog.run()
        dialog.destroy()

    def expertHandler (self, widget):
        self.destroy() # triggers main_quit
        complexui.start(self.configList)
        gtk.main()

    def configModified (self, node, b=True):
        if b != True:
            return
        # Save the configuration file
        try:
            file = open (self.userConfig.fileName, "w")
        except IOError:
            dialog = gtk.MessageDialog (
                commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
                _("Can't open \"%s\" for writing.") % self.userConfig.fileName)
            dialog.run()
            dialog.destroy()
            self.inConfigModified = False
            return
        self.commit()
        file.write (str(self.userConfig))
        file.close()

    def validateDriverPanel (self):
        self.validate()

def start (configList):
    userConfig = getUserConfig(configList)
    if not userConfig:
        dialog = gtk.MessageDialog (
            None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
            _("The DRI configuration file \"%s\" is broken or could not be "
              "created.") % os.path.join (os.environ["HOME"], ".drirc") +" "+
            _("DRIconf will be started in expert mode."))
        dialog.run()
        dialog.destroy()
        complexui.start(configList)
        return
    if not userConfig.writable:
        # Not writable: start expert mode
        dialog = gtk.MessageDialog (
            None, 0, gtk.MESSAGE_WARNING, gtk.BUTTONS_OK,
            _("Your DRI configuration file \"%s\" is not writable.") %
            userConfig.fileName +" "+
            _("DRIconf will be started in expert mode."))
        dialog.run()
        dialog.destroy()
        complexui.start(configList)
        return
    normalizedDeviceConfigs = normalizeConfig(configList, commonui.dpy)
    if normalizedDeviceConfigs == None:
        dialog = gtk.MessageDialog (
            None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
            _("Normalization of your DRI configuration file \"%s\" failed. "
              "Please report a bug with the original configuration file "
              "attached. The file will be treated as read-only for now.") %
            userConfig.fileName +" "+
            _("DRIconf will be started in expert mode."))
        dialog.run()
        dialog.destroy()
        userConfig.writable = False
        complexui.start(configList)
        return
    mainWindow = MainWindow(configList)
    commonui.mainWindow = mainWindow
    mainWindow.set_default_size (-1, 500)
    mainWindow.show()
    # Save modified normalized configuration before we start
    if hasattr(userConfig, "isModified") and userConfig.isModified:
        mainWindow.configModified(userConfig)



# Python interface to DRI configuration

# Copyright (C) 2003-2006  Felix Kuehling

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Contact: http://fxk.de.vu/

import os
import string
import re
import xml.parsers.expat

class Error (Exception):
    """ Base class for DRIError and XMLError """
    def __init__ (self, problem):
        self.problem = problem
    def __str__ (self):
        return self.problem

class DRIError (Error):
    """ Errors interfacing with xdriinfo """
    pass

class XMLError (Error):
    """ Errors in the DRI configuration data """
    pass

def XDriInfo (argStr, dpy = None):
    """ Call xdriinfo and raise DRIError on different failure conditions """
    if dpy != None:
        dpyStr = "-display " + dpy + " "
    else:
        dpyStr = ""
    infopipe = os.popen ("xdriinfo " + dpyStr + argStr, "r")
    driInfo = infopipe.read()
    result = infopipe.close()
    if result != None:
        signal = result & 0xff
        status = result >> 8
        if signal != 0:
            raise DRIError ("XDriInfo killed by signal " + signal + ".")
        elif status == 127:
            raise DRIError ("XDriInfo not found.")
        else:
            raise DRIError ("XDriInfo returned with non-zero exit code.")
    return driInfo

def StrToValue (str, type):
    """ Helper: convert str to given type.

    Raises an XMLError if str is not of the correct type. """
    try:
        if type == "int" or type == "enum":
            return int (str);
        elif type == "float":
            return float (str);
        else:
            if str == "true":
                return 1
            elif str == "false":
                return 0
            else:
                raise ValueError
    except ValueError:
        raise XMLError ("invalid value '" + str + "' for type '" + type + "'")

def ValueToStr (value, type):
    """ Helper: convert value of given type to string. """
    if type == "int" or type == "enum" or type == "float":
        return str(value)
    elif value:
        return "true"
    else:
        return "false"

def GetDesc (desc, preferredLangs):
    """ Helper: get a description with a list of language preferences.

    If the specified languages are not available then try english.
    If that doesn't work either, return any description.
    If there are no descriptions at all, return None. """
    for lang in preferredLangs:
        if desc.has_key (lang):
            return desc[lang]
    if desc.has_key ("en"):
        return desc["en"]
    if len(desc.values()) > 0:
        return desc.values()[0]
    return None

class Range:
    """ An interval """
    def __init__ (self, str, type):
        """ Parse str as a range.

        Raises an XMLError if str is not a legal range. """
        assert type == "int" or type == "enum" or type == "float"
        list = string.split (str, ":")
        if len (list) == 0 or len (list) > 2:
            raise XMLError ("Invalid range '" + str + "'")
        if len (list) >= 1:
            self.start = StrToValue (list[0], type)
        if len (list) == 2:
            self.end = StrToValue (list[1], type)
        else:
            self.end = self.start

    def __str__ (self):
        if self.start == self.end:
            return str(self.start)
        else:
            return str(self.start) + ":" + str(self.end)

    def empty (self):
        return self.start == self.end

class OptDesc:
    """ An option description in one language with enum values. """
    def __init__ (self, lang, text):
        self.lang = lang
        self.text = text
        self.enums = {}

    def __str__ (self):
        result = '<description lang="' + self.lang + '" text="' + self.text + \
                 '">\n'
        for value in sort(self.enums.keys()):
            result = result + '<enum value="' + str(value) + '" text="' + \
                     self.enums[value] + '" />\n'
        result = result + '</description>'
        return result

class OptInfo:
    """ All advertised information about an option. """
    def __init__ (self, name, type, default, valid = None):
        """ Initialize option information.

        Raises XMLError if
          - type is illegal
          - default is not a valid value of type
          - valid attribute is specified for a bool option
          - valid is specified but illegal """
        self.name = name

        if type != "int" and type != "enum" and type != "float" \
               and type != "bool":
            raise XMLError ("invalid type '" + type + "'")
        self.type = type
        self.valid = None
        if valid:
            if type == "bool":
                raise XMLError (
                    "valid attribute is not allowed with bool options")
            else:
                self.valid = [Range(x, type)
                              for x in string.split (valid, ",")]
        if not self.validate (default):
            raise XMLError ("default value is out of valid range")
        else:
            self.default = StrToValue (default, type)
        self.desc = {}

    def __str__ (self):
        result = '<option name="' + self.name + '" type="' + self.type + \
                 '" default="' + ValueToStr(self.default, self.type) + '" '
        if self.valid:
            return result + 'valid="' + \
                   reduce(lambda x,y: x+','+y, map(str,self.valid)) + \
                   '" />'
        else:
            return result + '/>'

    def validate (self, str):
        """ Check that str is of correct type and in a valid range. """
        try:
            v = StrToValue (str, self.type)
        except XMLError:
            return 0
        if self.valid:
            for r in self.valid:
                if v >= r.start and v <= r.end:
                    return 1
            return 0
        else:
            return 1

    def getDesc (self, preferredLangs):
        return GetDesc (self.desc, preferredLangs)

class OptSection:
    """ Representation of an option section.

    Contains descriptions and OptInfos as dictionaries. Options are also
    in a list so they can be extracted in a meaningful order. """
    def __init__ (self):
        """ Desc and options are initialized empty. """
        self.desc = {}
        self.options = {}
        self.optList = []

    def __str__ (self):
        result = '    <section>\n'
        for opt in self.options.values ():
            result = result + '        ' + str(opt) + '\n'
        result = result + '    </section>'
        return result

    def validate (self, valDict):
        """ Validate a dictionary of option values agains this OptSection. """
        allValid = 1
        for name,opt in self.options.items():
            if valDict.has_key (name):
                allValid = allValid and opt.validate (valDict[name])
        return allValid

    def getDesc (self, preferredLangs):
        return GetDesc (self.desc, preferredLangs)

class DriverInfo:
    """ Maintains a list of option sections and options in them. """
    def startElement (self, name, attr):
        """ Handle start_element events from XML parser. """
        if name == "section":
            self.curOptSection = OptSection ()
            self.optSections.append (self.curOptSection)
        elif name == "option":
            if self.curOptSection == None:
                raise XMLError ("option outside a section")
            if not attr.has_key ("name") or not attr.has_key ("type") or \
               not attr.has_key ("default"):
                raise XMLError ("mandatory option attribute missing")
            if attr.has_key ("valid"):
                self.curOption = OptInfo (attr["name"], attr["type"],
                                          attr["default"], attr["valid"])
            else:
                self.curOption = OptInfo (attr["name"], attr["type"],
                                          attr["default"])
            self.curOptSection.options[attr["name"]] = self.curOption
            self.curOptSection.optList.append (self.curOption)
        elif name == "description":
            if not attr.has_key ("lang") or not attr.has_key ("text"):
                raise XMLError ("description attribute missing")
            if self.curOption != None:
                self.curOptDesc = OptDesc(attr["lang"], attr["text"])
                self.curOption.desc[attr["lang"]] = self.curOptDesc
            elif self.curOptSection != None:
                self.curOptSection.desc[attr["lang"]] = attr["text"]
            else:
                raise XMLError ("description outside an option or section")
        elif name == "enum":
            if not attr.has_key ("value") or not attr.has_key ("text"):
                raise XMLError ("enum attribute missing")
            if self.curOptDesc != None:
                value = attr["value"]
                if not self.curOption.validate (value):
                    raise XMLError ("enum value is out of valid range")
                else:
                    value = StrToValue (value, self.curOption.type)
                self.curOptDesc.enums[value] = attr["text"]
            else:
                raise XMLError ("enum outside an option description")

    def endElement (self, name):
        """ Handle end_element events from XML parser. """
        if name == "section":
            self.curOptSection = None
        elif name == "option":
            self.curOption = None
        elif name == "description":
            self.curOptDesc = None

    def __init__ (self, name):
        """ Obtain and parse config info for this driver.

        Raises a DRIError if the driver does not support configuration.

        Raises a XMLError if the config info is illegal. """
        self.name = name
        driInfo = XDriInfo ("options " + name)

        self.optSections = []
        self.curOptSection = None
        self.curOption = None
        self.curOptDesc = None
        p = xml.parsers.expat.ParserCreate()

        p.StartElementHandler = self.startElement
        p.EndElementHandler = self.endElement

        try:
            p.Parse (driInfo)
        except xml.parsers.expat.ExpatError, problem:
            raise XMLError ("ExpatError: " + str(problem))

    def __str__ (self):
        result = '<driconf>\n'
        for sect in self.optSections:
            result = result + str(sect) + '\n'
        result = result + '</driconf>\n'
        return result

    def validate (self, valDict):
        """ Validate a dictionary of option values against this DriverInfo. """
        allValid = 1
        for optSection in self.optSections:
            allValid = allValid and optSection.validate (valDict)
        return allValid

    def getOptInfo (self, name):
        """ Return an option info for a given option name.

        If no such option exists in any section, None is returned. """
        for optSection in self.optSections:
            if optSection.options.has_key(name):
                return optSection.options[name]
        return None

class GLXInfo:
    def __init__ (self, screen, dpy):
        if dpy == None:
            if os.environ.has_key("DISPLAY"):
                dpy = os.environ["DISPLAY"]
            else:
                dpy = ":0"
        dot = dpy.find(".")
        if dot != -1:
            dpy = dpy[:dot]
        dpyStr = "-display " + dpy + "." + str(screen)
        infopipe = os.popen ("glxinfo " + dpyStr, "r")
        glxInfo = infopipe.read()
        result = infopipe.close()
        if result != None:
            signal = result & 0xff
            status = result >> 8
            if signal != 0:
                raise DRIError ("glxinfo killed by signal " + signal + ".")
            elif status == 127:
                raise DRIError ("glxinfo not found.")
            else:
                raise DRIError ("glxinfo returned with non-zero exit code.")
        else:
            # Parse
            vMatch = re.search ("^OpenGL vendor string: (.*)$", glxInfo, re.M)
            rMatch = re.search("^OpenGL renderer string: (.*)$", glxInfo, re.M)
            self.vendor = vMatch and vMatch.group(1)
            self.renderer = rMatch and rMatch.group(1)
            if not self.vendor or not self.renderer:
                raise DRIError ("unable to parse glxinfo output.")

class ScreenInfo:
    """ References a DriverInfo object with the real config info. """
    def __init__ (self, screen, dpy = None):
        """ Find or create the driver for this screen.

        Raises a DRIError if the screen is not direct rendering capable or
        if the DRI driver does not support configuration.

        Raises a XMLError if the config info is illegal. """
        self.num = screen
        driverName = XDriInfo ("driver " + str(screen), dpy)
        try:
            self.driver = GetDriver (driverName, 0)
        except XMLError, problem:
            raise XMLError (str(problem) + " (driver " + driverName + ")")
        else:
            try:
                self.glxInfo = GLXInfo (screen, dpy)
            except DRIError:
                self.glxInfo = None

class DisplayInfo:
    """ Maintains config info for all screens and drivers on a display """
    drivers = {}

    def __init__ (self, dpy = None):
        """ Find all direct rendering capable screens on dpy.

        Raises a DRIError if xdriinfo does not work for some reason. """
        self.dpy = dpy
        nScreens = int(XDriInfo ("nscreens", dpy))
        self.screens = [None for i in range (nScreens)]
        for i in range (nScreens):
            self.getScreen (i)

    def getScreen (self, i):
        """ Get the screen object for screen i.

        Returns None if the screen is not direct rendering capable or if
        the DRI driver does not support configuration.

        Raises a XMLError if the DRI driver's configuration information is
        invalid. """
        if i < 0 or i >= len(self.screens):
            return None
        if self.screens[i] != None:
            return self.screens[i]
        try:
            screen = ScreenInfo (i, self.dpy)
        except DRIError:
            screen = None
        except XMLError, problem:
            self.screens[i] = None
            raise XMLError (str(problem) + " (screen " + str(i) + ")")
        self.screens[i] = screen
        return screen

def GetDriver (name, catch=1):
    """ Get the driver object for the named driver.

    Returns None if the DRI driver does not support configuration.

    Raises a XMLError if the DRI driver's configuration information is
    invalid. """
    if DisplayInfo.drivers.has_key (name):
        return DisplayInfo.drivers[name]
    try:
        driver = DriverInfo (name)
    except DRIError, problem:
        if catch:
            driver = None
        else:
            raise DRIError (problem)
    else:
        DisplayInfo.drivers[name] = driver
    return driver

class AppConfig:
    """ Configuration data of an application given by the executable name.

    If no executable name is specified it applies to all applications. """
    def __init__ (self, device, name, executable = None):
        self.device = device
        self.name = name
        self.executable = executable
        self.options = {}

    def __str__ (self):
        result = '        <application name="' + self.name + '"'
        if self.executable != None:
            result = result + ' executable="' + self.executable + '">\n'
        else:
            result = result + '>\n'
        for n, v in self.options.items ():
            result = result + '            <option name="' + n + \
                     '" value="' + v + '" />\n'
        result = result + '        </application>'
        return result

class DeviceConfig:
    """ Configuration data of a device given by screen and/or driver.

    If neither screen nor driver is specified it applies to all devices. """
    def __init__ (self, config, screen = None, driver = None):
        self.config = config
        self.screen = screen
        self.driver = driver
        self.apps = []

    def __str__ (self):
        result = '    <device'
        if self.screen:
            result = result + ' screen="' + self.screen + '"'
        if self.driver:
            result = result + ' driver="' + self.driver + '"'
        result = result + '>\n'
        for a in self.apps:
            result = result + str(a) + '\n'
        result = result + '    </device>'
        return result

    def getDriver (self, display):
        """ Get the driver object for this device.

        Throws XMLError if the driver's config info is invalid. """
        driver = None
        if self.driver:
            driver = GetDriver (self.driver)
        elif self.screen:
            try:
                screenNum = int(self.screen)
            except ValueError:
                pass
            else:
                screen = display.getScreen(screenNum)
                if screen != None:
                    driver = screen.driver
        return driver

class DRIConfig:
    """ Configuration object representing one configuration file. """
    def startElement (self, name, attr):
        """ Handle start_element events from XML parser. """
        if name == "device":
            if attr.has_key ("screen") and attr.has_key ("driver"):
                self.curDevice = DeviceConfig (self, attr["screen"], attr["driver"])
            elif attr.has_key ("screen"):
                self.curDevice = DeviceConfig (self, screen = attr["screen"])
            elif attr.has_key ("driver"):
                self.curDevice = DeviceConfig (self, driver = attr["driver"])
            else:
                self.curDevice = DeviceConfig (self)
            self.devices.append (self.curDevice)
        elif name == "application":
            if self.curDevice == None:
                raise XMLError ("application outside a device")
            if not attr.has_key ("name"):
                raise XMLError ("mandatory application attribute missing")
            if attr.has_key ("executable"):
                self.curApp = AppConfig (self.curDevice, attr["name"],
                                         attr["executable"])
            else:
                self.curApp = AppConfig (self.curDevice, attr["name"])
            self.curDevice.apps.append (self.curApp)
        elif name == "option":
            if self.curApp == None:
                raise XMLError ("option outside an application")
            if not attr.has_key ("name") or not attr.has_key ("value"):
                raise XMLError ("option attribute missing")
            self.curApp.options[attr["name"]] = attr["value"]

    def endElement (self, name):
        """ Handle end_element events from XML parser. """
        if name == "device":
            self.curDevice = None
        elif name == "application":
            self.curApp = None
    
    def __init__ (self, file, fileName=""):
        """ Parse configuration file. """
        self.devices = []
        self.curDevice = None
        self.curApp = None
        if file:
            self.fileName = file.name
            p = xml.parsers.expat.ParserCreate()
            p.StartElementHandler = self.startElement
            p.EndElementHandler = self.endElement
            try:
                p.ParseFile (file)
            except xml.parsers.expat.ExpatError, problem:
                raise XMLError ("ExpatError: " + str(problem))
        else:
            self.fileName = fileName

    def __str__ (self):
        result = '<driconf>\n'
        for d in self.devices:
            result = result + str(d) + '\n'
        result = result + '</driconf>'
        return result



# DRI configuration GUI: common UI components and helpers, global vars

# Copyright (C) 2003-2006  Felix Kuehling

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Contact: http://fxk.de.vu/

import os
import locale
import gettext
import math
import dri
import pygtk
pygtk.require ("2.0")
import gtk
import gobject

# Install translations. Search in the current directory first (for
# easy testing). Then search in the default location and in
# /usr/local/share/locale. If all this fails fall back to the null
# translation.
try:
    _ = gettext.translation ("driconf", ".").ugettext
except IOError:
    try:
        _ = gettext.translation ("driconf").ugettext
    except IOError:
        _ = gettext.translation ("driconf", "/usr/local/share/locale",
                                 fallback=True).ugettext

# global variable: lang
locale.setlocale(locale.LC_ALL, '')
lang,encoding = locale.getlocale(locale.LC_MESSAGES)
if lang:
    underscore = lang.find ('_')
    if underscore != -1:
        lang = lang[0:underscore]
else:
    lang = "en"
# encoding is only a dummy. Pango uses UTF-8 everywhere. :)
del encoding

# global variable: version
version = "0.9.1"

# global variable: dpy
dpy = None

# global variable: main window
mainWindow = None

# Helper function:
# Find a file that should have been installed in .../shared/driconf
# Prefixes of __file__ are tried. And the current directory as a fallback.
def findInShared (name):
    # try all <prefix>/share/driconf/name for all prefixes of __file__
    head,tail = os.path.split (__file__)
    while head and tail:
        f = os.path.join (head, "share/driconf", name)
        if os.path.isfile (f):
            return f
        head,tail = os.path.split (head)
    # try name in the current directory
    if os.path.isfile (name):
        return name
    print "Warning: could not find %s." % name
    # nothing found
    return None

def fileIsWritable(filename):
    """ Find out if a file is writable.

    Returns True for existing writable files, False otherwise. """
    try:
        fd = os.open (filename, os.O_WRONLY)
    except OSError:
        return False
    if fd == -1:
        return False
    else:
        os.close (fd)
        return True

# Helper function:
# escape text that is going to be passed as markup to pango
def escapeMarkup (text):
    return text.replace ("<", "&lt;").replace (">", "&gt;")

class StockImage (gtk.Image):
    """ A stock image. """
    def __init__ (self, stock, size):
        """ Constructor. """
        gtk.Image.__init__ (self)
        self.set_from_stock (stock, size)

class WrappingLabel (gtk.Label):
    """ Line wrapping label with a highlight method """
    def __init__ (self, label, justify=gtk.JUSTIFY_LEFT, wrap=True,
                  width=-1, height=-1):
        self.text = escapeMarkup (label)
        gtk.Label.__init__ (self, label)
        self.set_justify (justify)
        self.set_line_wrap (wrap)
        self.set_size_request (width, height)

    def highlight (self, flag):
        """ Highlight the label. """
        if flag:
            self.set_markup ('<span foreground="red">' +  self.text + \
                             '</span>')
        else:
            self.set_markup (self.text)

class WrappingDummyCheckButton (gtk.HBox):
    """ Dummy Check button with a line wrapping label. """
    def __init__ (self, label, justify=gtk.JUSTIFY_LEFT, wrap=True,
                  width=-1, height=-1):
        """ Constructor. """
        gtk.HBox.__init__ (self)
        self.text = escapeMarkup (label)
        self.label = WrappingLabel (label, justify, wrap, width, height)
        self.label.show()
        self.pack_start (self.label, False, False, 0)

    def highlight (self, flag):
        """ Highlight the label. """
        self.label.highlight (flag)

    def get_active (self):
        """ Hack: make it behave like a check button ...

        ... to some extent such that OptionLine does not need to
        distinguish between simple and complex mode. """
        return True

class WrappingCheckButton (gtk.CheckButton):
    """ Check button with a line wrapping label. """
    def __init__ (self, label, justify=gtk.JUSTIFY_LEFT, wrap=True,
                  width=-1, height=-1):
        """ Constructor. """
        gtk.CheckButton.__init__ (self)
        checkHBox = gtk.HBox()
        self.text = escapeMarkup (label)
        self.label = WrappingLabel (label, justify, wrap, width, height)
        self.label.show()
        checkHBox.pack_start (self.label, False, False, 0)
        checkHBox.show()
        self.add (checkHBox)

    def highlight (self, flag):
        """ Highlight the label. """
        self.label.highlight (flag)

class WrappingOptionMenu (gtk.Button):
    """ Something that looks similar to a gtk.OptionMenu ...

    but can wrap the text and has a simpler interface. It acts as a
    bidirectional map from option descriptions (opt) to values (val)
    at the same time. """
    def __init__ (self, optValList, callback, justify=gtk.JUSTIFY_LEFT,
                  wrap=True, width=-1, height=-1):
        """ Constructor. """
        gtk.Button.__init__ (self)
        self.callback = callback
        self.optDict = {}
        self.valDict = {}
        for opt,val in optValList:
            self.optDict[opt] = val
            self.valDict[val] = opt
        firstOpt,firstVal = optValList[len(optValList)-1]
        self.value = firstVal
        hbox = gtk.HBox()
        arrow = gtk.Arrow (gtk.ARROW_DOWN, gtk.SHADOW_OUT)
        arrow.show()
        self.label = gtk.Label(firstOpt)
        self.label.set_justify (justify)
        self.label.set_line_wrap (wrap)
        self.label.set_size_request (width, height)
        self.label.show()
        hbox.pack_start (self.label, True, False, 5)
        hbox.pack_start (arrow, False, False, 0)
        hbox.show()
        self.add (hbox)
        self.menu = gtk.Menu()
        for opt,val in optValList:
            item = gtk.MenuItem (opt)
            item.connect("activate", self.menuSelect, opt)
            item.show()
            self.menu.append (item)
        self.connect("event", self.buttonPress)

    def setOpt (self, opt):
        """ Select an option by the option description. """
        self.label.set_text (opt)
        self.value = self.optDict[opt]

    def setValue (self, value):
        """ Select an option by its value. """
        self.label.set_text (self.valDict[value])
        self.value = value

    def getValue (self):
        """ Return the current value. """
        return self.value

    def buttonPress (self, widget, event):
        """ Popup the menu. """
        if event.type == gtk.gdk.BUTTON_PRESS:
            self.menu.popup(None, None, None, event.button, event.time)
            return True
        else:
            return False

    def menuSelect (self, widget, opt):
        """ React to selection of a menu item by the user. """
        self.setOpt (opt)
        self.callback (self)

class SlideSpinner (gtk.VBox):
    """ A spin button with a slider below.

    This is used for representing int and float options with a single
    range. The slider is only displayed on float options or on integer
    options with a large range (>= 20, arbitrary threshold). """
    def __init__ (self, callback, lower, upper, integer):
        gtk.VBox.__init__ (self)
        diff = upper - lower
        self.isInteger = integer
        if integer:
            step = 1
            page = (diff+4) / 5
            self.digits = 0
        else:
            self.digits = -int(math.floor(math.log10(diff) + 0.5)) + 3
            step = math.pow(10, -self.digits + 1)
            page = step * 10
            if self.digits < 0:
                self.digits = 0
        self.callback = callback
        self.adjustment = gtk.Adjustment (lower, lower, upper, step, page)
        self.spinner = gtk.SpinButton (self.adjustment, step, self.digits)
        self.spinner.set_numeric (True)
        self.spinner.show()
        self.pack_start (self.spinner, False, False, 0)
        if not integer or diff >= 20:
            self.slider = gtk.HScale (self.adjustment)
            self.slider.set_size_request (200, -1)
            self.slider.set_draw_value (False)
            self.slider.show()
            self.pack_start (self.slider, False, False, 0)
        self.adjConn = self.adjustment.connect ("value-changed", callback)

    def getValue (self):
        self.spinner.update()
        if self.isInteger:
            return int(self.adjustment.get_value())
        else:
            return self.adjustment.get_value()

    def setValue (self, value):
        self.adjustment.disconnect (self.adjConn)
        self.adjustment.set_value (value)
        self.adjConn = self.adjustment.connect ("value-changed", self.callback)

class OptionLine:
    """ One line in a SectionPage. """
    def __init__ (self, page, i, opt, simple, removable = False):
        """ Constructor. """
        self.page = page
        self.opt = opt
        self.index = i
        typeString = opt.type
        if opt.valid:
            typeString = typeString+" ["+ \
                         reduce(lambda x,y: x+','+y, map(str,opt.valid))+"]"
        # a check button with an option description
        desc = opt.getDesc([lang])
        if desc != None:
            desc = desc.text
        else:
            desc = u"(no description available)"
        if simple:
            self.label = WrappingDummyCheckButton (desc, width=200)
        else:
            self.label = WrappingCheckButton (desc, width=200)
            self.label.set_active (page.app.options.has_key (opt.name))
            self.label.set_sensitive (page.app.device.config.writable)
            self.label.connect ("clicked", self.checkOpt)
        tooltipString = str(opt)
        page.tooltips.set_tip (self.label, tooltipString)
        self.label.show()
        page.table.attach (self.label, 0, 1, i, i+1,
                           gtk.EXPAND|gtk.FILL, 0, 5, 5)
        # a button to reset the option to its default value
        sensitive = self.label.get_active() and page.app.device.config.writable
        self.resetButton = gtk.Button ()
        if removable:
            pixmap = StockImage ("gtk-remove", gtk.ICON_SIZE_SMALL_TOOLBAR)
        else:
            pixmap = StockImage ("gtk-undo", gtk.ICON_SIZE_SMALL_TOOLBAR)
        pixmap.show()
        self.resetButton.add (pixmap)
        self.resetButton.set_relief (gtk.RELIEF_NONE)
        self.resetButton.set_sensitive (sensitive)
        if removable:
            page.tooltips.set_tip(self.resetButton, _("Remove"))
            self.resetButton.connect ("clicked", self.removeOpt)
        else:
            page.tooltips.set_tip(self.resetButton, _("Reset to default value"))
            self.resetButton.connect ("clicked", self.resetOpt)
        self.resetButton.show()
        page.table.attach (self.resetButton, 2, 3, i, i+1, 0, 0, 5, 5)

        # get the option value, if it's invalid leave it as a string
        if page.app.options.has_key (opt.name):
            self.isValid = opt.validate (page.app.options[opt.name])
            try:
                value = dri.StrToValue (page.app.options[opt.name], opt.type)
            except dri.XMLError:
                self.isValid = False
            if not self.isValid:
                value = page.app.options[opt.name]
        else:
            value = opt.default
            self.isValid = True
        # the widget for editing the option value
        self.initWidget (opt, value)
        self.widget.set_sensitive (sensitive)
        page.table.attach (self.widget, 1, 2, i, i+1, gtk.FILL, 0, 5, 5)

    def initWidget (self, opt, value):
        """ Initialize the option widget.

        The widget type is selected automatically based in the option type
        and the set/range of valid values. """
        type = opt.type
        if not self.isValid:
            type = "invalid"
        if type == "bool":
            self.widget = gtk.ToggleButton ()
            self.widget.set_use_stock (True)
            if value:
                self.widget.set_label ("gtk-yes")
            else:
                self.widget.set_label ("gtk-no")
            self.widget.set_active (value)
            self.widget.connect ("toggled", self.activateSignal)
        elif type == "int" and opt.valid and len(opt.valid) == 1:
            self.widget = SlideSpinner (self.activateSignal, opt.valid[0].start,
                                        opt.valid[0].end, True)
            self.widget.setValue (value)
        elif type == "float" and opt.valid and len(opt.valid) == 1:
            self.widget = SlideSpinner (self.activateSignal, opt.valid[0].start,
                                        opt.valid[0].end, False)
            self.widget.setValue (value)
        elif type == "enum" or \
             (type != "invalid" and opt.valid and
              reduce (lambda x,y: x and y, map(dri.Range.empty, opt.valid))):
            desc = opt.getDesc([lang])
            optValList = []
            for r in opt.valid:
                if type == "enum":
                    for v in range (r.start, r.end+1):
                        vString = dri.ValueToStr(v, type)
                        if type == "enum" and desc and desc.enums.has_key(v):
                            string = desc.enums[v]
                        else:
                            string = vString
                        optValList.append ((string, vString))
                else:
                    vString = dri.ValueToStr(r.start, type)
                    optValList.append ((vString, vString))
            self.widget = WrappingOptionMenu (optValList, self.activateSignal,
                                              width=180)
            self.widget.setValue(str(value))
        else:
            self.widget = gtk.Entry ()
            if type == "invalid":
                self.widget.set_text (value)
            else:
                self.widget.set_text (dri.ValueToStr(value, type))
            self.widget.connect ("changed", self.activateSignal)
        self.highlightInvalid()
        self.widget.show()

    def updateWidget (self, value, valid):
        """ Update the option widget to a new value. """
        if self.widget.__class__ == gtk.ToggleButton:
            self.widget.set_active (value)
        elif self.widget.__class__ == SlideSpinner:
            self.widget.setValue (value)
        elif self.widget.__class__ == WrappingOptionMenu:
            self.widget.setValue(str(value))
        elif self.widget.__class__ == gtk.Entry:
            if self.opt.type == "bool" and valid:
                if value:
                    newText = "true"
                else:
                    newText = "false"
            else:
                newText = str(value)
            # only set new text if it changed, otherwise the changed signal
            # is triggered without a real value change
            if newText != self.widget.get_text():
                self.widget.set_text (newText)

    def getValue (self):
        """ Get the current value from the option widget.

        Returns None if the widget is not activated. """
        if not self.label.get_active():
            return None
        elif self.widget.__class__ == gtk.ToggleButton:
            if self.widget.get_active():
                return "true"
            else:
                return "false"
        elif self.widget.__class__ == SlideSpinner:
            return str(self.widget.getValue())
        elif self.widget.__class__ == WrappingOptionMenu:
            return self.widget.getValue()
        elif self.widget.__class__ == gtk.Entry:
            return self.widget.get_text()
        else:
            return None

    def checkOpt (self, widget):
        """ Handler for 'check button (maybe) toggled'. """
        if self.label.get_active():
            self.widget.set_sensitive (True)
            self.resetButton.set_sensitive (True)
        else:
            self.widget.set_sensitive (False)
            self.resetButton.set_sensitive (False)
        if not self.isValid:
            self.highlightInvalid()
            self.page.doValidate()
        self.page.optionModified (self)

    def activateSignal (self, widget, dummy=None):
        """ Handler for 'widget was activated by the user'. """
        if self.widget.__class__ == gtk.ToggleButton:
            value = self.widget.get_active()
            if value:
                self.widget.set_label ("gtk-yes")
            else:
                self.widget.set_label ("gtk-no")
        self.doValidate()
        self.page.optionModified (self)

    def resetOpt (self, widget):
        """ Reset to default value. """
        self.updateWidget (self.opt.default, True)
        self.page.optionModified (self)

    def removeOpt (self, widget):
        """ Reset to default value. """
        self.page.removeOption (self, self.opt)

    def doValidate (self):
        """ Validate the current value from the option widget.

        This is only interesting if the check button is active. Only
        gtk.Entry widgets should ever give invalid values in practice.
        Invalid option widgets are highlighted. If the validity
        changed, reinitialize the widget and have the page check if
        its validity was changed, too. """
        value = self.getValue()
        if value == None:
            return
        valid = self.opt.validate (value)
        if (valid and not self.isValid) or (not valid and self.isValid):
            self.isValid = valid
            self.highlightInvalid()
            # Re-init the widget.
            i = self.index
            self.page.table.remove(self.widget)
            self.widget.destroy()
            self.initWidget(self.opt, dri.StrToValue(value, self.opt.type))
            self.widget.show()
            self.page.table.attach (self.widget, 1, 2, i, i+1, gtk.FILL, 0, 5, 5)
            # Update parent page.
            self.page.doValidate()

    def highlightInvalid (self):
        self.label.highlight (not self.isValid and self.label.get_active())

    def validate (self):
        return self.isValid or not self.label.get_active()

class SectionPage (gtk.ScrolledWindow):
    """ One page in the DriverPanel with one OptionLine per option. """
    def __init__ (self, optSection, app, simple):
        """ Constructor. """
        gtk.ScrolledWindow.__init__ (self)
        self.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.optSection = optSection
        self.app = app
        self.simple = simple
        self.tooltips = gtk.Tooltips()
        self.table = gtk.Table (len(optSection.optList), 3)
        self.optLines = []
        for i in range (len(optSection.optList)):
            self.optLines.append (OptionLine (self, i, optSection.optList[i], simple))
        self.table.show()
        self.add_with_viewport (self.table)
        self.doValidate (init=True)

    def optionModified (self, optLine):
        """ Callback that is invoked by changed option lines. """
        self.app.modified(self.app)

    def doValidate (self, init=False):
        """ Validate the widget settings.

        The return value indicates if there are invalid option values. """
        valid = True
        for optLine in self.optLines:
            if not optLine.validate():
                valid = False
                break
        if not init and \
               ((valid and not self.isValid) or (not valid and self.isValid)):
            self.isValid = valid
            mainWindow.validateDriverPanel()
        self.isValid = valid

    def validate (self):
        return self.isValid

    def commit (self):
        """ Commit the widget settings. """
        for optLine in self.optLines:
            name = optLine.opt.name
            value = optLine.getValue()
            if value == None and self.app.options.has_key(name):
                del self.app.options[name]
            elif value != None:
                self.app.options[name] = value

class UnknownSectionPage(gtk.VBox):
    """ Special section page for options unknown to the driver. """
    def __init__ (self, driver, app):
        """ Constructor. """
        gtk.VBox.__init__ (self)
        scrolledWindow = gtk.ScrolledWindow ()
        scrolledWindow.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.app = app
        self.driver = driver
        self.nameEditable = app.device.config.writable and not driver
        self.valueEditable = app.device.config.writable
        # copy options (dict function does not exist in Python 2.1 :( )
        opts = {}
        for name,val in app.options.items():
            opts[name] = val
        # remove all options known to the driver
        self.driverOpts = {}
        if driver:
            for sect in driver.optSections:
                for opt in sect.optList:
                    self.driverOpts[opt.name] = opt
                    if opts.has_key (opt.name):
                        del opts[opt.name]
        # short cut
        self.opts = []
        if driver and len(opts) == 0:
            return
        # list all remaining options here
        self.store = gtk.ListStore (gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_BOOLEAN, gobject.TYPE_BOOLEAN)
        self.view = gtk.TreeView (self.store)
        self.view.set_rules_hint (True)
        optionRenderer = gtk.CellRendererText()
        optionRenderer.connect ("edited", self.editedSignal, 0)
        column = gtk.TreeViewColumn (_("Option"), optionRenderer,
                                     text=0, editable=2)
        self.view.append_column (column)
        valueRenderer = gtk.CellRendererText()
        valueRenderer.connect ("edited", self.editedSignal, 1)
        column = gtk.TreeViewColumn (_("Value"), valueRenderer,
                                     text=1, editable=3)
        self.view.append_column (column)
        self.view.get_selection().set_mode (gtk.SELECTION_MULTIPLE)
        for name,val in opts.items():
            self.store.set (self.store.append(),
                            0, str(name), 1, str(val),
                            2, self.nameEditable, 3, self.valueEditable)
            self.opts.append (name)
        self.view.show()
        scrolledWindow.add (self.view)
        scrolledWindow.show()
        self.pack_start (scrolledWindow, True, True, 0)
        buttonBox = gtk.HButtonBox()
        buttonBox.set_layout (gtk.BUTTONBOX_END)
        if not self.driver:
            newButton = gtk.Button (stock="gtk-add")
            newButton.connect ("clicked", self.newSetting)
            newButton.set_sensitive(app.device.config.writable)
            newButton.show()
            buttonBox.add (newButton)
        deleteButton = gtk.Button (stock="gtk-remove")
        deleteButton.connect ("clicked", self.deleteSelection)
        deleteButton.set_sensitive(app.device.config.writable)
        deleteButton.show()
        buttonBox.add (deleteButton)
        helpButton = gtk.Button (stock="gtk-help")
        helpButton.connect ("clicked", self.help)
        helpButton.show()
        buttonBox.add (helpButton)
        buttonBox.set_child_secondary (helpButton, True)
        buttonBox.show()
        self.pack_start (buttonBox, False, False, 0)

    def validate (self):
        """ These options can't be validated. """
        return True

    def commit (self):
        """ These options are never changed. """
        pass

    def deleteSelection (self, widget):
        """ Delete the selected items from the list and app config. """
        # Damn it! gtk.TreeSelection.get_selected_rows doesn't exist.
        # So we iterate over the whole list store and delete all selected
        # items.
        cur = self.store.get_iter_first()
        i = 0
        while cur:
            next = self.store.iter_next (cur)
            if self.view.get_selection().iter_is_selected (cur):
                name = self.store.get_value (cur, 0)
                del self.app.options[name]
                del self.opts[i]
                self.store.remove (cur)
                self.app.modified(self.app)
            else:
                i = i + 1
            cur = next

    def newSetting (self, widget):
        """ Create a new setting. Choose a unique option name, that is
        unknown to the driver. """
        name = "option"
        val = ""
        i = 0
        while self.app.options.has_key(name) or \
                  self.driverOpts.has_key(name):
            i = i + 1
            name = "option%d" % i
        self.app.options[name] = val
        self.opts.append (name)
        self.store.set (self.store.append(), 0, str(name), 1, str(val),
                        2, self.nameEditable, 3, self.valueEditable)
        self.app.modified(self.app)

    def help (self, widget):
        if self.driver:
            msg = _("This tab shows settings that are no longer supported "
                    "by the 3D driver. Unless you're using the same "
                    "configuration file on another computer with a different "
                    "driver version, it is safe to remove these settings.")
        else:
            msg = _("The driver for this device could not be determined or "
                    "does not support configuration. You can still change "
                    "the settings but it cannot be verified whether they are "
                    "supported and valid.")
        dialog = gtk.MessageDialog (
            mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_INFO, gtk.BUTTONS_CLOSE, msg)
        dialog.connect ("response", lambda d,r: d.destroy())
        dialog.show()

    def editedSignal (self, widget, row, newVal, value):
        row = int(row)
        name = self.opts[row]
        cursor = self.store.get_iter_first()
        for i in range(row): cursor = self.store.iter_next (cursor)
        if value:
            if self.app.options[name] == newVal:
                return
            self.app.options[name] = newVal
            self.store.set (cursor, 0, str(name), 1, str(newVal),
                            2, self.nameEditable, 3, self.valueEditable)
        else:
            if name == newVal or self.app.options.has_key(newVal) or \
                   self.driverOpts.has_key(newVal):
                return
            val = self.app.options.pop(name)
            name = newVal
            self.opts[row] = name
            self.app.options[name] = val
            self.store.set (cursor, 0, str(name), 1, str(val),
                            2, self.nameEditable, 3, self.valueEditable)
        self.app.modified(self.app)

if gtk.__dict__.has_key("AboutDialog"):
    # About Dialog was added in gtk 2.6.
    class AboutDialog (gtk.AboutDialog):
        def __init__ (self):
            gtk.AboutDialog.__init__(self)
            translators = _("translator-credits")
            if translators == "translator-credits":
                translators = None
            self.set_name("DRIconf")
            self.set_version(version)
            self.set_copyright(u"Copyright \u00a9 2003-2006  "
                               u"Felix K\u00fchling")
            self.set_comments(_("A configuration applet for DRI drivers"))
            self.set_website(u"http://dri.freedesktop.org/wiki/DriConf")
            if translators:
                self.set_translator_credits(translators)
            logoPath = findInShared("drilogo.jpg")
            if logoPath:
                logo = gtk.gdk.pixbuf_new_from_file (logoPath)
                self.set_logo(logo)
else:
    class AboutDialog (gtk.MessageDialog):
        def __init__ (self):
            translators = _("translator-credits")
            if translators == "translator-credits":
                translators = None
            text = u"DRIconf %s\n" \
                   u"%s\n" \
                   u"Copyright \u00a9 2003-2005  Felix K\u00fchling\n" \
                   u"\n" \
                   u"http://dri.freedesktop.org/wiki/DriConf" \
                   % (version,  _("A configuration applet for DRI drivers"))
            if translators:
                text = text + (u"\n\n%s: %s" % (_("Translated by"),
                                              _("translator-credits")))
            gtk.MessageDialog.__init__(
                self, mainWindow,
                gtk.DIALOG_DESTROY_WITH_PARENT|gtk.DIALOG_MODAL,
                gtk.MESSAGE_INFO, gtk.BUTTONS_CLOSE, text)
            self.set_title(_("About DRIconf"))



# DRI configuration GUI: complex UI components

# Copyright (C) 2003-2006  Felix Kuehling

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Contact: http://fxk.de.vu/

import dri
import pygtk
pygtk.require ("2.0")
import gtk
import gobject

import driconf_commonui
commonui = driconf_commonui    # short cut

from driconf_commonui import _, lang, findInShared, escapeMarkup, WrappingCheckButton, SectionPage, UnknownSectionPage

class DriverPanel (gtk.Frame):
    """ Panel for driver settings for a specific application. """
    def __init__ (self, driver, app):
        """ Constructor. """
        gtk.Frame.__init__ (self)
        frameLabel = gtk.Label()
        frameLabel.set_markup ("<b>" + escapeMarkup(
            _("Application")+": "+app.name) + "</b>")
        frameLabel.show()
        self.set_label_widget (frameLabel)
        self.driver = driver
        self.app = app
        tooltips = gtk.Tooltips()
        table = gtk.Table(2, 2)
        self.execCheck = WrappingCheckButton (_("Apply only to this executable"))
        self.execCheck.set_sensitive (app.device.config.writable)
        tooltips.set_tip (self.execCheck, _(
            "Leave this disabled to configure all applications.\n"
            "Beware that some applications or games are just a shell script "
            "that starts a real executable with a different name."))
        self.execCheck.show()
        table.attach (self.execCheck, 0, 1, 0, 1, 0, 0, 5, 5)
        self.execEntry = gtk.Entry()
        if app.executable != None:
            self.execCheck.set_active (True)
            self.execEntry.set_text (app.executable)
        self.execEntry.set_sensitive (app.device.config.writable and
                                      app.executable != None)
        self.execEntry.show()
        self.execCheck.connect ("toggled", self.execToggled)
        self.execEntry.connect ("changed", self.execChanged)
        table.attach (self.execEntry, 1, 2, 0, 1,
                      gtk.EXPAND|gtk.FILL, 0, 5, 5)
        notebook = gtk.Notebook()
        notebook.popup_enable()
        notebook.set_scrollable (True)
        self.sectPages = []
        self.sectLabels = []
        unknownPage = UnknownSectionPage (driver, app)
        if not driver or len(unknownPage.opts) > 0:
            unknownPage.show()
            unknownLabel = gtk.Label (_("Unknown"))
            unknownLabel.show()
            notebook.append_page (unknownPage, unknownLabel)
            self.sectPages.append (unknownPage)
            self.sectLabels.append (unknownLabel)
        if driver:
            for sect in driver.optSections:
                sectPage = SectionPage (sect, app, False)
                sectPage.show()
                desc = sect.getDesc([lang])
                if desc:
                    sectLabel = gtk.Label (desc)
                    sectLabel.set_line_wrap (True)
                else:
                    sectLabel = gtk.Label (_("(no description)"))
                sectLabel.show()
                notebook.append_page (sectPage, sectLabel)
                self.sectPages.append (sectPage)
                self.sectLabels.append (sectLabel)
        if len(self.sectLabels) > 0:
            style = self.sectLabels[0].get_style()
            self.default_normal_fg = style.fg[gtk.STATE_NORMAL].copy()
            self.default_active_fg = style.fg[gtk.STATE_ACTIVE].copy()
        self.validate()
        notebook.show()
        table.attach (notebook, 0, 2, 1, 2,
                      gtk.FILL, gtk.EXPAND|gtk.FILL, 5, 5)
        table.show()
        self.add (table)

    def execChanged (self, widget):
        self.app.modified(self.app)

    def execToggled (self, widget):
        self.execEntry.set_sensitive (self.execCheck.get_active())
        self.app.modified(self.app)

    def validate (self):
        """ Validate the configuration.

        Labels of invalid section pages are highlighted. Returns whether
        there were invalid option values. """
        index = 0
        allValid = True
        for sectPage in self.sectPages:
            valid = sectPage.validate()
            if not valid:
                # strange, active and normal appear to be swapped :-/
                self.sectLabels[index].modify_fg (
                    gtk.STATE_NORMAL, gtk.gdk.Color (65535, 0, 0))
                self.sectLabels[index].modify_fg (
                    gtk.STATE_ACTIVE, gtk.gdk.Color (65535, 0, 0))
            else:
                self.sectLabels[index].modify_fg (
                    gtk.STATE_NORMAL, self.default_normal_fg)
                self.sectLabels[index].modify_fg (
                    gtk.STATE_ACTIVE, self.default_active_fg)
            allValid = allValid and valid
            index = index+1
        return allValid

    def commit (self):
        """ Commit changes to the configuration. """
        executable = self.execEntry.get_text()
        if not self.execCheck.get_active():
            if self.app.executable != None:
                self.app.executable = None
                self.app.modified(self.app)
        elif executable != self.app.executable:
            self.app.executable = executable
            self.app.modified(self.app)
        for sectPage in self.sectPages:
            sectPage.commit()

    def renameApp (self):
        """ Change the application name. """
        self.set_label ("Application: " + self.app.name)

class NameDialog (gtk.Dialog):
    """ Dialog for setting the name of an application. """
    def __init__ (self, title, callback, name, data):
        gtk.Dialog.__init__ (self, title, commonui.mainWindow,
                             gtk.DIALOG_DESTROY_WITH_PARENT|gtk.DIALOG_MODAL,
                             (gtk.STOCK_OK, gtk.RESPONSE_OK,
                              gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        self.callback = callback
        self.data = data
        self.connect ("response", self.responseSignal)
        table = gtk.Table(2, 2)
        commentLabel = gtk.Label (_(
            "Enter the name of the application below. This just serves as "
            "a description for you. Don't forget to set the executable "
            "afterwards."))
        commentLabel.set_line_wrap (True)
        commentLabel.show()
        table.attach (commentLabel, 0, 2, 0, 1,
                      gtk.EXPAND|gtk.FILL, gtk.EXPAND, 10, 5)
        label = gtk.Label (_("Application Name"))
        label.show()
        table.attach (label, 0, 1, 1, 2, 0, gtk.EXPAND, 10, 5)
        self.entry = gtk.Entry()
        self.entry.set_text (name)
        self.entry.select_region (0, len(name))
        self.entry.connect ("activate", self.activateSignal)
        self.entry.show()
        table.attach (self.entry, 1, 2, 1, 2,
                      gtk.EXPAND|gtk.FILL, gtk.EXPAND, 10, 5)
        table.show()
        self.vbox.pack_start (table, True, True, 5)
        self.show()
        self.entry.grab_focus()

    def activateSignal (self, widget):
        self.response (gtk.RESPONSE_OK)

    def responseSignal (self, dialog, response):
        if response == gtk.RESPONSE_OK:
            self.callback (self.entry.get_text(), self.data)
        self.destroy()

class DeviceDialog (gtk.Dialog):
    """ Dialog for choosing driver and screen of a device. """
    def __init__ (self, title, callback, data):
        gtk.Dialog.__init__ (self, title, commonui.mainWindow,
                             gtk.DIALOG_DESTROY_WITH_PARENT|gtk.DIALOG_MODAL,
                             (gtk.STOCK_OK, gtk.RESPONSE_OK,
                              gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        self.callback = callback
        self.data = data
        self.connect ("response", self.responseSignal)
        table = gtk.Table (2, 3)
        commentLabel = gtk.Label (_(
            "Describe the device that you would like to configure."))
        commentLabel.set_line_wrap (True)
        commentLabel.show()
        table.attach (commentLabel, 0, 2, 0, 1,
                      gtk.EXPAND|gtk.FILL, gtk.EXPAND, 10, 5)
        screenLabel = gtk.Label (_("Screen Number"))
        screenLabel.show()
        table.attach (screenLabel, 0, 1, 1, 2, 0, gtk.EXPAND, 10, 5)
        self.screenCombo = gtk.Combo()
        self.screenCombo.set_popdown_strings (
            [""]+map(str,range(len(commonui.dpy.screens))))
        self.screenCombo.entry.connect ("activate", self.screenSignal)
        self.screenCombo.list.connect ("select_child", self.screenSignal)
        self.screenCombo.show()
        table.attach (self.screenCombo, 1, 2, 1, 2,
                      gtk.EXPAND|gtk.FILL, gtk.EXPAND, 10, 5)
        driverLabel = gtk.Label (_("Driver Name"))
        driverLabel.show()
        table.attach (driverLabel, 0, 1, 2, 3, 0, gtk.EXPAND, 10, 5)
        self.driverCombo = gtk.Combo()
        self.driverCombo.set_popdown_strings (
            [""]+[str(driver.name) for driver in dri.DisplayInfo.drivers.values()])
        self.driverCombo.show()
        table.attach (self.driverCombo, 1, 2, 2, 3,
                      gtk.EXPAND|gtk.FILL, gtk.EXPAND, 10, 5)
        if data and data.__class__ == dri.DeviceConfig:
            if data.screen:
                self.screenCombo.entry.set_text (data.screen)
            if data.driver:
                self.driverCombo.entry.set_text (data.driver)
        table.show()
        self.vbox.pack_start (table, True, True, 5)
        self.show()

    def screenSignal (self, widget, data=None):
        screenName = self.screenCombo.entry.get_text()
        try:
            screenNum = int(screenName)
        except ValueError:
            pass
        else:
            screen = commonui.dpy.getScreen(screenNum)
            if screen != None:
                driver = screen.driver
                self.driverCombo.entry.set_text (str(driver.name))

    def responseSignal (self, dialog, response):
        if response == gtk.RESPONSE_OK:
            self.callback (self.screenCombo.entry.get_text(),
                           self.driverCombo.entry.get_text(), self.data)
        self.destroy()

class ConfigTreeModel (gtk.GenericTreeModel):
    # constructur
    def __init__ (self, configList):
        gtk.GenericTreeModel.__init__ (self)
        self.configList = []
        for config in configList:
            self.addNode (config)

    def iconFromShared (self, name, size, missing):
        path = findInShared (name)
        if path:
            icon = gtk.gdk.pixbuf_new_from_file (path)
            return icon.scale_simple (size, size, gtk.gdk.INTERP_BILINEAR)
        else:
            return missing

    def renderIcons (self, widget):
        self.configIcon = widget.render_icon ("gtk-properties",
                                              gtk.ICON_SIZE_MENU, None)
        self.appIcon = widget.render_icon ("gtk-execute",
                                           gtk.ICON_SIZE_MENU, None)
        self.unspecIcon = widget.render_icon ("gtk-dialog-question",
                                              gtk.ICON_SIZE_MENU, None)
        missing = widget.render_icon ("gtk-missing-image",
                                      gtk.ICON_SIZE_MENU, None)
        size = max (missing.get_width(), missing.get_height())
        self.screenIcon = self.iconFromShared ("screen.png", size, missing)
        self.driverIcon = self.iconFromShared ("card.png", size, missing)
        self.screenDriverIcon = self.iconFromShared ("screencard.png",
                                                     size, missing)

    # implementation of the GenericTreeModel interface
    def on_get_flags (self):
        return gtk.TREE_MODEL_ITERS_PERSIST
    def on_get_n_columns (self):
        return 2
    def on_get_column_type (self, col):
        if col == 0:
            return gobject.TYPE_OBJECT
        else:
            return gobject.TYPE_STRING
    def on_get_path (self, node):
        if node.__class__ == dri.DRIConfig:
            return (self.configList.index(node),)
        elif node.__class__ == dri.DeviceConfig:
            config = node.config
            return (self.configList.index(config),
                    config.devices.index(node))
        elif node.__class__ == dri.AppConfig:
            device = node.device
            config = device.config
            return (self.configList.index(config),
                    config.devices.index(device),
                    device.apps.index(node))
        else:
            assert 0
    def on_get_iter (self, path):
        if len(path) > 3:
            return None
        configIndex = path[0] # config
        if configIndex < len(self.configList):
            config = self.configList[configIndex]
        else:
            return None
        if len(path) == 1:
            return config
        deviceIndex = path[1] # device
        if deviceIndex < len(config.devices):
            device = config.devices[deviceIndex]
        else:
            return None
        if len(path) == 2:
            return device
        appIndex = path[2] # application
        if appIndex < len(device.apps):
            app = device.apps[appIndex]
        else:
            return None
        return app
    def on_get_value (self, node, col):
        if node.__class__ == dri.DRIConfig:
            if col == 0:
                return self.configIcon
            else:
                return "<b>" + escapeMarkup(str(node.fileName)) + "</b>"
        elif node.__class__ == dri.DeviceConfig:
            if node.screen and node.driver:
                name = _("%s on screen %s") % (node.driver.capitalize(),
                                               node.screen)
                icon = self.screenDriverIcon
            elif node.screen:
                name = _("Screen %s") % node.screen
                icon = self.screenIcon
            elif node.driver:
                name = "%s" % node.driver.capitalize()
                icon = self.driverIcon
            else:
                name = _("Unspecified device")
                icon = self.unspecIcon
            if col == 0:
                return icon
            else:
                return escapeMarkup(str(name))
        elif node.__class__ == dri.AppConfig:
            if col == 0:
                return self.appIcon
            elif not node.isValid:
                return '<span foreground="red">' + \
                       escapeMarkup(str(node.name)) + '</span>'
            else:
                return escapeMarkup(str(node.name))
    def on_iter_next (self, node):
        if node.__class__ == dri.DRIConfig:
            list = self.configList
        elif node.__class__ == dri.DeviceConfig:
            list = node.config.devices
        elif node.__class__ == dri.AppConfig:
            list = node.device.apps
        else:
            return None
        index = list.index(node)
        if index+1 < len(list):
            return list[index+1]
        else:
            return None
    def on_iter_children (self, node):
        return self.on_iter_nth_child (node, 0)
    def on_iter_has_child (self, node):
        return (self.on_iter_n_children (node) > 0)
    def on_iter_n_children (self, node):
        if not node:
            return len(self.configList)
        elif node.__class__ == dri.DRIConfig:
            return len(node.devices)
        elif node.__class__ == dri.DeviceConfig:
            return len(node.apps)
        else:
            return 0
    def on_iter_nth_child (self, node, index):
        if not node:
            list = self.configList
        elif node.__class__ == dri.DRIConfig:
            list = node.devices
        elif node.__class__ == dri.DeviceConfig:
            list = node.apps
        else:
            return None
        if len(list) > index:
            return list[index]
        else:
            return None
    def on_iter_parent (self, node):
        if node.__class__ == dri.DeviceConfig:
            return node.config
        elif node.__class__ == dri.AppConfig:
            return node.device
        else:
            return None

    # helpers for converting between nodes, paths and TreeIterators
    def getPathFromNode (self, node):
        return self.on_get_path (node)
    def getNodeFromPath (self, path):
        return self.on_get_iter (path)
    def getIterFromNode (self, node):
        return self.get_iter (self.getPathFromNode(node))

    # config list
    def getConfigList (self):
        return self.configList

    # callback for registring modifications
    def nodeModified (self, node, b=True):
        if node.__class__ == dri.DRIConfig:
            config = node
        elif node.__class__ == dri.DeviceConfig:
            config = node.config
        elif node.__class__ == dri.AppConfig:
            config = node.device.config
        if config.isModified != b:
            config.isModified = b
            curNode = commonui.mainWindow.configTree.getSelection (allowNone=True)
            if not curNode:
                commonui.mainWindow.deactivateButtons()
            elif curNode.__class__ == dri.DRIConfig:
                commonui.mainWindow.activateConfigButtons(curNode)
            elif curNode.__class__ == dri.DeviceConfig:
                commonui.mainWindow.activateDeviceButtons(curNode)
            elif curNode.__class__ == dri.AppConfig:
                commonui.mainWindow.activateAppButtons(curNode)

    # higher level model manipulations
    def addNode (self, node, sibling = None):
        """ Add a new node and inform the TreeView. """
        self.initNode (node)
        if node.__class__ == dri.DRIConfig:
            list = self.configList
        elif node.__class__ == dri.DeviceConfig:
            list = node.config.devices
        elif node.__class__ == dri.AppConfig:
            list = node.device.apps
        if sibling != None:
            index = list.index (sibling)
            list.insert (index, node)
        else:
            list.append (node)
        self.registerNode (node)
    def initNode (self, node):
        node.modified = self.nodeModified
        if node.__class__ == dri.DRIConfig:
            if not hasattr(node,"isModified"):
                node.isModified = False
            for device in node.devices:
                self.initNode (device)
        elif node.__class__ == dri.DeviceConfig:
            for app in node.apps:
                self.initNode (app)
        elif node.__class__ == dri.AppConfig:
            self.validateAppNode (node)
    def registerNode (self, node):
        path = self.on_get_path (node)
        iter = self.get_iter (path)
        self.row_inserted (path, iter)
        if node.__class__ == dri.DRIConfig:
            for device in node.devices:
                self.registerNode (device)
        elif node.__class__ == dri.DeviceConfig:
            for app in node.apps:
                self.registerNode (app)
    def removeNode (self, node):
        if node.__class__ == dri.DRIConfig:
            list = self.configList
            while len(node.devices) > 0:
                self.removeNode (node.devices[0])
        elif node.__class__ == dri.DeviceConfig:
            list = node.config.devices
            while len(node.apps) > 0:
                self.removeNode (node.apps[0])
        elif node.__class__ == dri.AppConfig:
            list = node.device.apps
        path = self.on_get_path (node)
        list.remove (node)
        self.row_deleted (path)

    # find the first writable application
    def findFirstWritableApp (self):
        foundApp = None
        for config in self.configList:
            if not config.writable:
                continue
            for device in config.devices:
                if len(device.apps) > 0:
                    foundApp = device.apps[0]
                    break
        return foundApp

    # validate an application node
    def validateAppNode (self, app):
        try:
            driver = app.device.getDriver(commonui.dpy)
        except dri.XMLError:
            driver = None
        if driver and not driver.validate (app.options):
            app.isValid = False
        else:
            app.isValid = True

class ConfigTreeView (gtk.TreeView):
    def __init__ (self, configList):
        self.model = ConfigTreeModel (configList)
        gtk.TreeView.__init__ (self, self.model)
        self.model.renderIcons (self)
        self.connect ("style-set", lambda widget, prevStyle:
                      self.model.renderIcons (widget))
        self.set_size_request (200, -1)
        self.set_headers_visible (False)
        self.expand_all()
        self.get_selection().set_mode (gtk.SELECTION_BROWSE)
        self.get_selection().connect ("changed", self.selectionChangedSignal)
        column = gtk.TreeViewColumn()
        column.set_spacing (2)
        renderPixbuf = gtk.CellRendererPixbuf()
        column.pack_start (renderPixbuf, expand=False)
        column.add_attribute (renderPixbuf, "pixbuf", 0)
        renderText = gtk.CellRendererText()
        column.pack_start (renderText, expand=True)
        column.add_attribute (renderText, "markup", 1)
        self.append_column (column)

    def getConfigList (self):
        return self.model.getConfigList()

    # selection handling
    def getSelection (self, allowNone=False):
        model, iter = self.get_selection().get_selected()
        assert iter or allowNone
        if iter:
            path = self.model.get_path (iter)
            return self.model.getNodeFromPath (path)
        else:
            return None
    def selectFirstWritableApp (self):
        app = self.model.findFirstWritableApp()
        if app:
            path = self.model.getPathFromNode (app)
            self.get_selection().select_path (path)
            self.scroll_to_cell (path=path, use_align=False)
    def selectionChangedSignal (self, data):
        node = self.getSelection (allowNone=True)
        if not node:
            driver = None
            app = None
        elif node.__class__ == dri.AppConfig:
            app = node
            try:
                driver = app.device.getDriver (commonui.dpy)
            except dri.XMLError, problem:
                driver = None
                dialog = gtk.MessageDialog (
                    commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
                    gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
                    _("Parsing the driver's configuration information: %s") %
                    problem)
                dialog.connect ("response", lambda d,r: d.destroy())
                dialog.show()
        else:
            driver = None
            app = None
        commonui.mainWindow.commitDriverPanel()
        commonui.mainWindow.switchDriverPanel (driver, app)
        if not node:
            commonui.mainWindow.deactivateButtons()
        elif node.__class__ == dri.DRIConfig:
            commonui.mainWindow.activateConfigButtons(node)
        elif node.__class__ == dri.DeviceConfig:
            commonui.mainWindow.activateDeviceButtons(node)
        elif node.__class__ == dri.AppConfig:
            commonui.mainWindow.activateAppButtons(node)

    # highlight invalid nodes
    def validateAppNode (self, app):
        self.model.validateAppNode (app)
        # Emit a row_changed event so that the view gets updated properly.
        path = self.model.getPathFromNode (app)
        iter = self.model.get_iter (path)
        self.model.row_changed (path, iter)

    # UI actions on the config tree
    def moveUp (self, widget):
        self.moveItem (-1)
    def moveDown (self, widget):
        self.moveItem (1)
    def removeItem (self, widget):
        node = self.getSelection()
        if node.__class__ == dri.AppConfig:
            parent = node.device
            dialog = gtk.MessageDialog (
                commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO,
                _("Really delete application \"%s\"?") % node.name)
        elif node.__class__ == dri.DeviceConfig:
            parent = node.config
            dialog = gtk.MessageDialog (
                commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO,
                _("Really delete device and all applications in it?"))
        else:
            # The remove button should be unsensitive.
            assert False
        response = dialog.run()
        dialog.destroy ()
        if response != gtk.RESPONSE_YES:
            return
        if node.__class__ == dri.AppConfig:
            commonui.mainWindow.removeApp (node)
        elif node.__class__ == dri.DeviceConfig:
            for app in node.apps:
                commonui.mainWindow.removeApp (app)
        self.model.removeNode (node)
        parent.modified(parent)
        path = self.model.getPathFromNode (parent)
        self.get_selection().select_path (path)
        self.scroll_to_cell (path=path, use_align=False)
    def properties (self, widget):
        node = self.getSelection()
        if node.__class__ == dri.AppConfig:
            dialog = NameDialog (_("Rename Application"),
                                 self.renameCallback, node.name, node)
        else:
            dialog = DeviceDialog (_("Device Properties"),
                                   self.propertiesCallback, node)
    def newItem (self, widget):
        node = self.getSelection()
        if node.__class__ == dri.AppConfig:
            node = node.device
        if node.__class__ == dri.DeviceConfig:
            dialog = NameDialog (_("New Application"), self.newAppCallback, "",
                                 node)
        elif node.__class__ == dri.DRIConfig:
            dialog = DeviceDialog (_("New Device"), self.newDeviceCallback,
                                   node)
    def saveConfig (self, widget):
        node = self.getSelection()
        if node.__class__ == dri.AppConfig:
            config = node.device.config
        elif node.__class__ == dri.DeviceConfig:
            config = node.config
        elif node.__class__ == dri.DRIConfig:
            config = node
        valid = True
        for device in config.devices:
            try:
                driver = device.getDriver (commonui.dpy)
            except dri.XMLError:
                driver = None
            if driver == None:
                continue
            for app in device.apps:
                valid = valid and driver.validate (app.options)
        if not valid:
            dialog = gtk.MessageDialog (
                commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO,
                _("The configuration contains invalid entries. Save anyway?"))
            response = dialog.run()
            dialog.destroy()
            if response != gtk.RESPONSE_YES:
                return
        try:
            file = open (config.fileName, "w")
        except IOError:
            dialog = gtk.MessageDialog (
                commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
                _("Can't open \"%s\" for writing.") % config.fileName)
            dialog.run()
            dialog.destroy()
            return
        commonui.mainWindow.commitDriverPanel()
        file.write (str(config))
        file.close()
        config.modified(config, False)
    def reloadConfig (self, widget):
        node = self.getSelection()
        if node.__class__ == dri.AppConfig:
            config = node.device.config
        elif node.__class__ == dri.DeviceConfig:
            config = node.config
        else:
            config = node
        dialog = gtk.MessageDialog (
            commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO,
            _("Really reload \"%s\" from disk?") % config.fileName)
        response = dialog.run()
        dialog.destroy()
        if response != gtk.RESPONSE_YES:
            return
        try:
            cfile = open (config.fileName, "r")
        except IOError:
            dialog = gtk.MessageDialog (
                commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
                _("Couldn't open \"%s\" for reading. "
                  "The file was not reloaded.") % config.fileName)
            dialog.run()
            dialog.destroy()
            return
        # Try to parse the configuration file.
        try:
            newConfig = dri.DRIConfig (cfile)
        except dri.XMLError, problem:
            dialog = gtk.MessageDialog (
                commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
                gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
                _("Configuration file \"%s\" contains errors:\n"
                  "%s\n"
                  "The file was not reloaded.") %
                (config.fileName, str(problem)))
            dialog.run()
            dialog.destroy()
            cfile.close()
            return
        cfile.close()
        # Check if the file is writable in the end.
        newConfig.writable = commonui.fileIsWritable (config.fileName)
        # find the position of config
        configList = self.getConfigList()
        index = configList.index (config)
        if index == len(configList)-1:
            sibling = None
        else:
            sibling = configList[index+1]
        if node.__class__ == dri.AppConfig:
            commonui.mainWindow.removeApp (node)
        elif node.__class__ == dri.DeviceConfig:
            for app in node.apps:
                commonui.mainWindow.removeApp (app)
        self.model.removeNode (config)
        self.model.addNode (newConfig, sibling)
        path = self.model.getPathFromNode (newConfig)
        self.expand_row (path, True)
        self.get_selection().select_path (path)
        self.scroll_to_cell (path=path, use_align=False)

    # helper function for moving tree nodes around
    def moveItem (self, inc):
        node = self.getSelection()
        if node.__class__ == dri.AppConfig:
            parent = node.device
            siblings = parent.apps
        elif node.__class__ == dri.DeviceConfig:
            parent = node.config
            siblings = parent.devices
        else:
            assert False
        index = siblings.index (node)
        newIndex = index+inc
        if newIndex < 0 or newIndex >= len(siblings):
            return
        siblings.remove (node)
        siblings.insert (newIndex, node)
        newOrder = range(len(siblings))
        newOrder[index] = newIndex
        newOrder[newIndex] = index
        path = self.model.getPathFromNode (parent)
        iter = self.model.get_iter (path)
        self.model.rows_reordered (path, iter, newOrder)
        parent.modified(parent)
        path = self.model.getPathFromNode (node)
        self.scroll_to_cell (path=path, use_align=False)

    # callbacks from dialogs
    def renameCallback (self, name, app):
        app.name = name
        path = self.model.getPathFromNode (app)
        iter = self.model.get_iter (path)
        self.model.row_changed (path, iter)
        commonui.mainWindow.renameApp (app)
        app.modified(app)
    def propertiesCallback (self, screen, driver, device):
        device.screen = screen
        device.driver = driver
        path = self.model.getPathFromNode (device)
        iter = self.model.get_iter (path)
        self.model.row_changed (path, iter)
        device.modified(device)
    def newAppCallback (self, name, device):
        app = dri.AppConfig (device, name)
        self.model.addNode (app)
        if len(device.apps) == 1:
            self.expand_row (self.model.getPathFromNode(device), True)
        device.modified(device)
        path = self.model.getPathFromNode (app)
        self.get_selection().select_path (path)
        self.scroll_to_cell (path=path, use_align=False)
    def newDeviceCallback (self, screen, driver, config):
        device = dri.DeviceConfig (config, screen, driver)
        self.model.addNode (device)
        config.modified(config)
        path = self.model.getPathFromNode (device)
        self.get_selection().select_path (path)
        self.scroll_to_cell (path=path, use_align=False)

class MainWindow (gtk.Window):
    """ The main window consiting of ConfigTree, DriverPanel and toolbar. """
    def __init__ (self, configList):
        gtk.Window.__init__ (self)
        self.set_title ("DRIconf")
        self.connect ("destroy", lambda dummy: gtk.main_quit())
        self.connect ("delete_event", self.exitHandler)
        self.vbox = gtk.VBox()
        self.paned = gtk.HPaned()
        self.configTree = ConfigTreeView (configList)
        self.configTree.show()
        scrolledWindow = gtk.ScrolledWindow ()
        scrolledWindow.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolledWindow.add (self.configTree)
        scrolledWindow.show()
        self.paned.add1(scrolledWindow)
        self.paned.show()
        self.toolbar = gtk.Toolbar ()
        iconSize = self.toolbar.get_icon_size()
        self.saveButton = self.toolbar.insert_stock (
            "gtk-save", _("Save selected configuration file"),
            "priv", self.configTree.saveConfig, None, -1)
        self.reloadButton = self.toolbar.insert_stock (
            "gtk-revert-to-saved", _("Reload selected configuration file"),
            "priv", self.configTree.reloadConfig, None, -1)
        self.toolbar.append_space()
        self.newButton = self.toolbar.insert_stock (
            "gtk-new", _("Create a new device or application"),
            "priv", self.configTree.newItem, None, -1)
        self.removeButton = self.toolbar.insert_stock (
            "gtk-delete", _("Remove selected device or application"),
            "priv", self.configTree.removeItem, None, -1)
        self.upButton = self.toolbar.insert_stock (
            "gtk-go-up", _("Move selected item up"),
            "priv", self.configTree.moveUp, None, -1)
        self.downButton = self.toolbar.insert_stock (
            "gtk-go-down", _("Move selected item down"),
            "priv", self.configTree.moveDown, None, -1)
        self.propertiesButton = self.toolbar.insert_stock (
            "gtk-properties", _("Properties of selected device or application"),
            "priv", self.configTree.properties, None, -1)
        self.toolbar.append_space()
        # The gtk-about stock item is available with gtk >= 2.6.
        # It's definitely not available with gtk 2.2. Not sure about 2.4.
        if gtk.gtk_version[0] == 2 and gtk.gtk_version[1] < 6:
            aboutStock = "gtk-dialog-info"
        else:
            aboutStock = "gtk-about"
        self.aboutButton = self.toolbar.insert_stock (
            aboutStock, _("About DRIconf"), "priv",
            self.aboutHandler, None, -1)
        self.toolbar.append_space()
        self.exitButton = self.toolbar.insert_stock (
            "gtk-quit", _("Exit DRIconf"), "priv",
            self.exitHandler, None, -1)
        if len(configList) != 0:
            self.activateConfigButtons (configList[0])
        self.toolbar.show()
        self.vbox.pack_start (self.toolbar, False, True, 0)
        self.vbox.pack_start (self.paned, True, True, 0)
        self.vbox.show()
        self.add (self.vbox)
        self.curDriverPanel = None
        self.logo = gtk.EventBox ()
        logoPath = findInShared("drilogo.jpg")
        if logoPath:
            image = gtk.Image()
            image.set_from_file (logoPath)
            self.logo.add (image)
        self.logo.modify_bg (gtk.STATE_NORMAL,
                             gtk.gdk.Color (65535, 65535, 65535))
        self.logo.show_all()
        self.paned.add2 (self.logo)

    def initSelection (self):
        self.configTree.selectFirstWritableApp()

    def commitDriverPanel (self):
        if self.curDriverPanel != None:
            self.curDriverPanel.commit()
            self.configTree.validateAppNode (self.curDriverPanel.app)

    def validateDriverPanel (self):
        if self.curDriverPanel != None:
            self.curDriverPanel.validate()
            self.curDriverPanel.commit()
            self.configTree.validateAppNode (self.curDriverPanel.app)

    def switchDriverPanel (self, driver=None, app=None):
        if self.curDriverPanel != None:
            if self.curDriverPanel.driver == driver and \
               self.curDriverPanel.app == app:
                return
            self.paned.remove (self.curDriverPanel)
        elif app != None:
            self.paned.remove (self.logo)
        if app != None:
            self.curDriverPanel = DriverPanel (driver, app)
            self.curDriverPanel.show ()
            self.paned.add2 (self.curDriverPanel)
        elif self.curDriverPanel != None:
            self.curDriverPanel = None
            self.logo.show()
            self.paned.add2 (self.logo)

    def removeApp (self, app):
        if self.curDriverPanel != None and self.curDriverPanel.app == app:
            self.paned.remove (self.curDriverPanel)
            self.curDriverPanel = None
            self.logo.show()
            self.paned.add2 (self.logo)

    def renameApp (self, app):
        if self.curDriverPanel != None and self.curDriverPanel.app == app:
            self.curDriverPanel.renameApp()

    def deactivateButtons (self):
        self.saveButton      .set_sensitive (False)
        self.reloadButton    .set_sensitive (False)
        self.newButton       .set_sensitive (False)
        self.removeButton    .set_sensitive (False)
        self.upButton        .set_sensitive (False)
        self.downButton      .set_sensitive (False)
        self.propertiesButton.set_sensitive (False)

    def activateConfigButtons (self, config):
        writable = config.writable
        modified = config.isModified
        self.saveButton      .set_sensitive (writable and modified)
        self.reloadButton    .set_sensitive (True)
        self.newButton       .set_sensitive (writable)
        self.removeButton    .set_sensitive (False)
        self.upButton        .set_sensitive (False)
        self.downButton      .set_sensitive (False)
        self.propertiesButton.set_sensitive (False)

    def activateDeviceButtons (self, device):
        writable = device.config.writable
        modified = device.config.isModified
        self.saveButton      .set_sensitive (writable and modified)
        self.reloadButton    .set_sensitive (True)
        self.newButton       .set_sensitive (writable)
        self.removeButton    .set_sensitive (writable)
        self.upButton        .set_sensitive (writable)
        self.downButton      .set_sensitive (writable)
        self.propertiesButton.set_sensitive (writable)

    def activateAppButtons (self, app):
        # Button sensitivity is identical for apps and devices.
        self.activateDeviceButtons (app.device)

    def aboutHandler (self, widget):
        dialog = commonui.AboutDialog()
        dialog.connect("response",
                       lambda dialog, response: dialog.destroy())
        dialog.show()

    def exitHandler (self, widget, event=None):
        modified = False
        for config in self.configTree.getConfigList():
            if config.isModified:
                modified = True
                break
        if modified:
            dialog = gtk.MessageDialog (
                commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT|gtk.DIALOG_MODAL,
                gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO,
                _("There are unsaved modifications. Exit anyway?"))
            dialog.connect ("response", self.doExit)
            dialog.show()
            return True
        elif event == None:
            # called from toolbar button: main_quit!
            gtk.main_quit()
        else:
            # called from delete_event: indicate it's ok to destroy
            return False

    def doExit (self, dialog, response):
        dialog.destroy()
        if response == gtk.RESPONSE_YES:
            gtk.main_quit()

def start (configList):
    # initSelection must be called before and after mainWindow.show().
    # Before makes sure that the initial window size is correct.
    # After is needed since the selection seems to get lost in
    # mainWindow.show().
    mainWindow = MainWindow(configList)
    commonui.mainWindow = mainWindow
    mainWindow.set_default_size (750, 375)
    mainWindow.initSelection()
    mainWindow.show ()
    mainWindow.initSelection()



from distutils.core import setup

langs = ["de", "es", "it", "ru", "nl"]
translations = []
for lang in langs:
    translations.append (("share/locale/%s/LC_MESSAGES" % lang,
                          ["%s/LC_MESSAGES/driconf.mo" % lang]))

setup(name="driconf",
      version="0.9.1",
      description="A configuration applet for DRI drivers",
      author="Felix Kuehling",
      author_email="fxkuehl@gmx.de",
      url="http://dri.freedesktop.org/wiki/DriConf",
      py_modules=["dri", "driconf", "driconf_commonui", "driconf_complexui",
                  "driconf_simpleui"],
      scripts=["driconf"],
      data_files=[("share/driconf", ["card.png", "screen.png", "screencard.png",
                                     "drilogo.jpg", "driconf-icon.png"])
                  ] + translations)

#
# Search for obsolete files.
#
# driconf_xpm is gone for good, the other two python modules were moved to
# <prefix>/lib/driconf/...
#
from os.path import isfile, isdir, join
from distutils.sysconfig import get_python_lib
obsoleteFiles = []
for prefix in [None, "/usr/local"]:
    if not prefix:
        pyLibPath = get_python_lib()
    else:
        pyLibPath = get_python_lib(prefix=prefix)
    if not isdir (pyLibPath):
        continue
    for f in ["driconf_xpm.py", "driconf_xpm.pyc", "driconf_xpm.pyo",
              "driconf.py", "driconf.pyc", "driconf.pyo",
              "dri.py", "dri.pyc", "dri.pyo"]:
        path = join (pyLibPath, f)
        if isfile (path):
            obsoleteFiles.append (path)
if obsoleteFiles:
    print "\n*** Obsolete files from previous DRIconf versions were found on " \
          "your system.\n*** Unless you tweaked setup.cfg you can probably " \
          "delete them:"
    for f in obsoleteFiles:
        print "***\t%s" % f

#
# Check if required packages are installed
#
errors = 0
try:
    import pygtk
    pygtk.require ("2.0")
    import gtk
except:
    print "\n*** Warning: importing GTK version 2 doesn't work."
    errors = 1
else:
    if gtk.check_version(2, 4, 0):
        print "\n*** Warning: DRIconf requires GTK 2.4 or newer."
        errors = 1
try:
    import xml.parsers.expat
except:
    if not errors:
        print
    print "*** Warning: importing xml.parsers.expat doesn't work."
    errors = 1

if errors:
    print "*** Warning: DRIconf will probably not work for the above reason(s)."



# DRI configuration GUI using python-gtk

# Copyright (C) 2003-2006  Felix Kuehling

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Contact: http://fxk.de.vu/

import sys
import os
import dri
import pygtk
pygtk.require ("2.0")
import gtk
import gobject

if gtk.check_version(2, 4, 0):
    print "Error: DRIconf requires GTK 2.4 or newer."
    sys.exit(1)

import driconf_commonui
import driconf_complexui
import driconf_simpleui

commonui = driconf_commonui     # short cuts
complexui = driconf_complexui
simpleui = driconf_simpleui

from driconf_commonui import _


def main():
    if len(sys.argv) >= 2 and sys.argv[1] == "-e":
        expert = True
    else:
        expert = False

    # read configuration information from the drivers
    try:
        commonui.dpy = dri.DisplayInfo ()
    except dri.DRIError, problem:
        dialog = gtk.MessageDialog (None, 0, gtk.MESSAGE_ERROR,
                                    gtk.BUTTONS_OK, str(problem))
        dialog.run()
        dialog.destroy()
        return
    except dri.XMLError, problem:
        dialog = gtk.MessageDialog (
            None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
            _("There are errors in a driver's configuration information:\n"
              "%s\n"
              "This should not happen. It probably means that you have to "
              "update DRIconf.") % str(problem))
        dialog.run()
        dialog.destroy()
        return

    configScreens = [screen for screen in commonui.dpy.screens
                     if screen != None]
    if len(configScreens) == 0:
        dialog = gtk.MessageDialog (
            None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
            _("Could not detect any configurable direct-rendering capable "
              "devices.")+" "+_("DRIconf will be started in expert mode."))
        dialog.run()
        dialog.destroy()
        expert = True

    # read or create configuration files
    fileNameList = ["/etc/drirc", os.path.join (os.environ["HOME"], ".drirc")]
    configList = []
    newFiles = []
    for fileName in fileNameList:
        try:
            cfile = open (fileName, "r")
        except IOError:
            # Make a default configuration file.
            config = dri.DRIConfig (None, fileName)
            config.writable = True
            for screen in configScreens:
                device = dri.DeviceConfig (config, str(screen.num),
                                           screen.driver.name)
                app = dri.AppConfig (device, "all")
                device.apps.append (app)
                config.devices.append (device)
            # Try to write the new file. If it fails, don't add this config.
            try:
                file = open (config.fileName, "w")
            except IOError:
                config = None
            else:
                file.write (str(config))
                file.close()
                newFiles.append (fileName)
        else:
            # Try to parse the configuration file.
            try:
                config = dri.DRIConfig (cfile)
            except dri.XMLError, problem:
                dialog = gtk.MessageDialog (
                    None, 0, gtk.MESSAGE_WARNING, gtk.BUTTONS_OK,
                    _("Configuration file \"%s\" contains errors:\n"
                      "%s\n"
                      "I will leave the file alone until you fix the problem "
                      "manually or remove the file.") %
                    (fileName, str(problem)))
                dialog.run()
                dialog.destroy()
                continue
            else:
                # Check if the file is writable in the end.
                config.writable = commonui.fileIsWritable (fileName)
            cfile.close()
        if config:
            configList.append (config)

    if len(configList) == 0:
        dialog = gtk.MessageDialog (
            None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
            _("There are no usable DRI configuration files and a new one "
              "could not be created. Exiting now."))
        dialog.run()
        dialog.destroy()
        return

    # open the main window
    if expert:
        complexui.start(configList)
    else:
        simpleui.start(configList)

    if len(newFiles) == 1:
        dialog = gtk.MessageDialog (
            commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_INFO, gtk.BUTTONS_OK,
            _("Created a new DRI configuration file \"%s\" for you.")
            % newFiles[0])
        dialog.run()
        dialog.destroy()
    elif len(newFiles) > 1:
        dialog = gtk.MessageDialog (
            commonui.mainWindow, gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_INFO, gtk.BUTTONS_OK,
            _("Created new DRI configuration files %s for you.") %
            reduce(lambda a, b: str(a) + ", " + str(b),
                   map (lambda a: "\"%s\"" % str(a), newFiles)))
        dialog.run()
        dialog.destroy()

    # run
    gtk.main()
